from __future__ import annotations

import dataclasses
import re
import typing as t
from pathlib import Path

from nr.util.generic import T
from nr.util.singleton import NotSet


@t.overload
def match_version_ref_pattern(filename: Path, pattern: str) -> VersionRef:
    ...


@t.overload
def match_version_ref_pattern(filename: Path, pattern: str, fallback: T) -> T | VersionRef:
    ...


def match_version_ref_pattern(filename: Path, pattern: str, fallback: NotSet | T = NotSet.Value) -> T | VersionRef:
    """Matches a regular expression in the given file and returns the location of the match. The *pattern*
    should contain at least one capturing group. The first capturing group is considered the one that contains
    the version number exactly.

    Arguments:
      filename: The file of which the contents will be checked against the pattern.
      pattern: The regular expression that contains at least one capturing group.
    """

    compiled_pattern = re.compile(pattern, re.M | re.S)
    if not compiled_pattern.groups:
        raise ValueError(
            f"pattern must contain at least one capturing group (filename: {filename!r}, pattern: {pattern!r})"
        )

    with open(filename) as fp:
        match = compiled_pattern.search(fp.read())
        if match:
            return VersionRef(filename, match.start(1), match.end(1), match.group(1), match.group(0))

    if fallback is not NotSet.Value:
        return fallback
    raise ValueError(f"pattern {pattern!r} does not match in file {filename!r}")


def match_version_ref_pattern_on_lines(filename: Path, pattern: str) -> list[VersionRef]:
    """Like #match_version_ref_pattern(), but returns all matches, but matches it on a line-by-line basis. The
    *pattern* must have a `version` group. The pattern is compiled with #re.M and #re.S flags."""

    compiled_pattern = re.compile(pattern, re.M | re.S)
    refs = []
    for match in re.finditer(compiled_pattern, filename.read_text()):
        refs.append(
            VersionRef(
                file=filename,
                start=match.start("version"),
                end=match.end("version"),
                value=match.group("version"),
                content=match.group(0),
            )
        )
    return refs


@dataclasses.dataclass
class VersionRef:
    """Represents a reference to a version number in a file."""

    file: Path
    start: int
    end: int
    value: str
    content: str

    def __post_init__(self) -> None:
        assert isinstance(self.file, Path), self



import typing as t
from pathlib import Path

from slap.ext.application.changelog import get_changelog_manager
from slap.plugins import ReleasePlugin
from slap.repository import Repository


class ChangelogReleasePlugin(ReleasePlugin):
    """Renames the `_unreleased.toml` file when a release is created."""

    def create_release(self, repository: Repository, target_version: str, dry: bool) -> t.Sequence[Path]:
        changed_files: list[Path] = []
        for project in repository.projects():
            manager = get_changelog_manager(repository, project)
            unreleased = manager.unreleased()
            new_version = manager.version(target_version)
            if unreleased.exists():
                cwd = Path.cwd()
                old = unreleased.path.relative_to(cwd)
                new = new_version.path.relative_to(cwd)
                if not changed_files:
                    self.io.write_line("releasing changelog")
                self.io.write_line(f"  <fg=cyan>{old}</fg> → <b>{new}</b>")
                if not dry:
                    unreleased.release(target_version)
                changed_files += [unreleased.path, new_version.path]
        return changed_files



import logging
import os
import shutil
import string
import subprocess as sp
import typing as t
from pathlib import Path

from nr.python.environment.virtualenv import VirtualEnvInfo, get_current_venv

from slap.application import Application, Command, argument, option
from slap.plugins import ApplicationPlugin
from slap.python.environment import PythonEnvironment

logger = logging.getLogger(__name__)

GLOBAL_BIN_DIRECTORY = Path("~/.local/bin").expanduser()
GLOBAL_VENVS_DIRECTORY = Path("~/.local/venvs").expanduser()
SHADOW_INIT_SCRIPTS = {
    "bash": """
    function slap() {
      local ORIGINAL=$(which slap)
      if ! [ $? = 0 ]; then
        >&2 echo "error: command 'slap' does not exist"
        return 127
      fi
      if [ "$1" == "venv" ] && [[ "$2" =~ -[gc]*a[gc]* ]]; then
        eval "$(SLAP_SHADOW=true "$ORIGINAL" "$@")"
      else
        "$ORIGINAL" "$@"
      fi
      return $?
    }
  """,
}

USER_INIT_SCRIPTS = {
    "bash": 'which slap >/dev/null && eval "$(SLAP_SHADOW=true slap venv -i bash)"',
}


class Venv(VirtualEnvInfo):
    def create(self, python_bin: str) -> None:
        self.path.parent.mkdir(parents=True, exist_ok=True)
        sp.check_call([python_bin, "-m", "venv", self.path])

    def delete(self) -> None:
        shutil.rmtree(self.path)


class VenvManager:
    def __init__(self, directory: Path | None = None) -> None:
        self.directory = directory or Path(".venvs")
        self.state_file = self.directory / ".state"

    def _get_state(self) -> dict[str, t.Any]:
        import json

        return json.loads(self.state_file.read_text()) if self.state_file.exists() else {}

    def _set_state(self, state: dict[str, t.Any]) -> None:
        import json

        self.state_file.write_text(json.dumps(state))

    def ls(self) -> t.Iterable[Venv]:
        if not self.directory.exists():
            return
        for path in self.directory.iterdir():
            if path.is_dir() and not path.name.startswith("."):
                yield Venv(path)

    def get(self, venv_name: str) -> Venv:
        return Venv(self.directory / venv_name)

    def get_last_activated(self) -> Venv | None:
        venv_name = self._get_state().get("last_active_environment")
        venv = self.get(venv_name) if venv_name else None
        return venv if venv and venv.exists() else None

    def set_last_activated(self, venv_name: str) -> None:
        state = self._get_state()
        state["last_active_environment"] = venv_name
        self._set_state(state)


class VenvAwareCommand(Command):
    """Base class for commands that should be aware of the active local virtual environment. Before the
    command is executed, it will check if we're currently in a virtual environment. If not, it will activate
    the environment that is considered "active" by the Slap `venv` command."""

    requires_venv: t.ClassVar[bool] = True

    options = [
        option(
            "no-venv-check",
            description="Do not check if the target Python environment is a virtual environment.",
        ),
        option(
            "ignore-active-venv",
            description="Ignore the currently active VIRTUAL_ENV and act as if it isn't set.",
        ),
        option(
            "use-venv",
            description="Use the specified Slap-managed virtual environment. This can be used to run a command "
            "in a seperate environment without activating or setting it as the current environment. Note that this "
            "option can not be used to run in a globally managed environment.",
            flag=False,
        ),
    ]

    def handle(self) -> int:
        if self.option("no-venv-check"):
            return 0
        venv = get_current_venv(os.environ)
        if (self.option("ignore-active-venv") or self.option("use-venv")) and venv:
            self.line(f"<info>(venv-aware) deactivating current virtual environment (<s>{venv.path}</s>)</info>")
            venv.deactivate(os.environ)
            venv = None
        if venv:
            self.io.error_output.write_line(
                "<info>(venv-aware) a virtual environment is already activated "
                f'(<s>{os.environ["VIRTUAL_ENV"]}</s>)</info>'
            )
        else:
            manager = VenvManager()
            if self.option("use-venv"):
                venv = manager.get(self.option("use-venv"))
                if not venv.exists():
                    self.io.error_output.write_line(f'<error>environment <s>"{venv.name}"</s> does not exist</error>')
                    return 1
            else:
                venv = manager.get_last_activated()
            if venv:
                venv.activate(os.environ)
                self.io.error_output.write_line(
                    f'<info>(venv-aware) activating current environment <s>"{venv.name}"</s></info>'
                )
            elif self.requires_venv:
                self.io.error_output.write_line(
                    "<warning>(venv-aware) there is no current environment that can be activated</warning>"
                )
                return 1
        return 0


class VenvCommand(Command):
    """Create, activate and remove virtual environments.

    This command makes it easy to create and manage virtual environments, locally as well
    as globally. Local environments are stored in the `.venvs/` directory in the current
    directory. Global environments are stored in `~/.local/venvs`.

    In order to be able to use the <opt>-a,--activate</opt> option directly from this command,
    it must be shadowed by a function in your shell. Use the <opt>-i,--init-code</opt> command
    to get a code snippet that you can place in your shell's init scripts.

    <u>Usage Example:</u>

        <code>$ slap venv -i bash >> ~/.profile; source ~/.profile
        $ slap venv -cg craftr
        <info>creating global environment <s>"craftr"</s> (using <code>python3</code>)</info>
        $ slap venv -lg</code>
        <info>1 environment in <s>"/home/niklas/.local/venvs"</s></info>
        • craftr   <code>3.10.2 (main, Jan 15 2022, 18:02:07) [GCC 9.3.0]</code>
        <code>$ slap venv -ag craftr
        (craftr) $ </code>

    Note that most Slap commands support using the active virtual environment it
    to be active in your shell (such as `slap run`, `slap test`, `slap install`,
    etc.).
    """

    name = "venv"
    arguments = [
        argument(
            "name",
            description="The environment name.",
            optional=True,
        ),
    ]
    options = [
        option(
            "--global",
            "-g",
            description="Manage virtual environments in the global scope instead of the local directory.",
        ),
        option(
            "--activate",
            "-a",
            description="Activate the environment given by the environment name. Note that using this option if used "
            "directly with the Slap CLI will cause an error because it needs to be shadowed by a function of your "
            "shell.",
        ),
        option(
            "--create",
            "-c",
            description="Create the environment with the given environment name. If no <opt>name</opt> is specified, "
            "the environment name will be the major.minor version of the current Python version.",
        ),
        option(
            "--delete",
            "-d",
            description="Delete the environment with the given environment name.",
        ),
        option(
            "--set",
            "-s",
            description="Similar to <opt>-a,--activate</opt>, but it will not activate the environment in your active "
            "shell even if you have the Slap shim installed (see <opt>-i,--init-code</opt>).",
        ),
        option(
            "--list",
            "-l",
            description="List the available environments.",
        ),
        option(
            "--path",
            "-p",
            description="Print the path of the specified or the current venv. Exit with status code 1 and no output if "
            "the environment does not exist or there is no current environment.",
        ),
        option("--exists", "-e", description="Return 0 if the specified environment exists, 1 otherwise."),
        option(
            "--init-code",
            "-i",
            description="Print the code snippet that can be placed in your shells init scripts to shadow this command "
            "in order to properly make use of the <opt>-a,--activate</opt> option. Currently supported shells are: "
            + ", ".join(USER_INIT_SCRIPTS),
            flag=False,
        ),
        option(
            "--python",
            description="The Python executable to use to create the virtual environment. If this is not specified, "
            "it defaults to <code>python</code> + the environment name if the environment name looks like a version "
            "number (contains numbers and dots). Otehrwise, it defaults to <code>python3</code>.",
            flag=False,
        ),
    ]

    def _validate_args(self) -> bool:
        for opt in ("activate", "create", "delete", "set", "exists"):
            if self.option("init-code") and self.option(opt):
                self.line_error(
                    f"error: <opt>-i,--init-code</opt> is not compatible with <opt>-{opt[0]},--{opt}</opt>", "error"
                )
                return False
            if self.option("list") and self.option(opt):
                self.line_error(
                    f"error: <opt>-l,--list</opt> is not compatible with <opt>-{opt[0]},--{opt}</opt>", "error"
                )
                return False
        for opt in ("delete", "set", "exists"):
            if self.option(opt) and not self.argument("name"):
                self.line_error("error: missing <opt>name</opt> argument", "error")
                return False
        for opt in ("activate", "create", "set", "exists"):
            if self.option("delete") and self.option(opt):
                self.line_error("error: <opt>-d,--delete</opt> is not compatible with <opt>--{opt}</opt>", "error")
                return False
        if self.option("path"):
            for opt in ("activate", "create", "delete", "set", "list", "init-code", "python", "exists"):
                if self.option(opt):
                    self.line_error("error: <opt>--path,-P</opt> is not compatible with <opt>--{opt}</opt>", "error")
                    return False
        if not any(
            self.option(opt) for opt in ("activate", "create", "delete", "set", "list", "init-code", "path", "exists")
        ):
            self.line_error("error: no operation specified", "error")
            return False
        return True

    def _get_python_bin(self) -> str:
        from slap.ext.application.install import get_active_python_bin

        python = get_active_python_bin(self, False)
        name = self.argument("name")
        if name and not python and set(name).issubset(string.digits + "."):
            python = f"python{name}"
        return python or "python3"

    def _list_environments(self, manager: VenvManager) -> None:
        venvs = list(manager.ls())
        if not venvs:
            self.line_error(f'no environments in <s>"{manager.directory}"</s>', "info")
            return
        self.line(f'{len(venvs)} environment{"s" if len(venvs) != 1 else ""} in <s>"{manager.directory}"</s>', "info")
        maxw = max(len(venv.name) for venv in venvs)
        for venv in venvs:
            self.line(f"• {venv.name.ljust(maxw)}  <code>{venv.get_python_version().splitlines()[0]}</code>")

    def _is_called_from_shadow(self) -> bool:
        return os.getenv("SLAP_SHADOW") == "true"

    def _get_init_code(self, shell: str) -> int:
        import textwrap

        source = SHADOW_INIT_SCRIPTS if self._is_called_from_shadow() else USER_INIT_SCRIPTS
        if shell in source:
            print(textwrap.dedent(source[shell]))
            return 0
        else:
            self.line_error(f"error: init code for shell <s>{shell}</s> is not supported", "error")
            return 1

    def _pick_venv(self, location: str, manager: VenvManager) -> Venv | None:
        """Picks either the last virtual environment that was activated, or the only one that is available."""

        venv = manager.get_last_activated()
        if venv and venv.exists():
            return venv

        venvs = list(manager.ls())
        if not venvs:
            self.line_error(f"error: no {location} environments", "error")
            return None

        if len(venvs) == 1:
            return venvs[0]

        self.line_error(f"error: multiple {location} environments exist, not sure which to pick", "error")
        return None

    def handle(self) -> int:
        if not self._validate_args():
            return 1

        shell = self.option("init-code")
        if shell:
            return self._get_init_code(shell)

        manager = VenvManager(GLOBAL_VENVS_DIRECTORY if self.option("global") else Path(".venvs"))

        if self.option("list"):
            self._list_environments(manager)
            return 0

        python = self._get_python_bin()
        venv = manager.get(self.argument("name")) if self.argument("name") else None
        location = "global" if self.option("global") else "local"

        if self.option("create"):
            if not venv:
                venv = manager.get(".".join(map(str, PythonEnvironment.of(python).version_tuple[:2])))
            if venv.exists():
                self.line_error(f'error: environment <s>"{venv.name}"</s> already exists', "error")
                return 1
            self.line_error(
                f'creating {location} environment <s>"{venv.name}"</s> (using <code>{python}</code>)', "info"
            )
            venv.create(python)

        if self.option("activate"):
            if not venv:
                venv = self._pick_venv(location, manager)
                if not venv:
                    return 1

            if not venv.exists():
                self.line_error(f'error: environment <s>"{venv.name}"</s> does not exist', "error")
                return 1

            if self._is_called_from_shadow():
                self.line_error(f'activating {location} environment <s>"{venv.name}"</s>', "info")
            else:
                self.line_error(
                    "warning: the <opt>-a,--activate</opt> option only works when shadowed by a shell function",
                    "warning",
                )

            # TODO (@NiklasRosenstein): Adjust output based on the shell that this is called from?
            # TODO (@NiklasRosenstein): Must be activate.cmd on Windows
            print(f'source "{venv.get_bin("activate")}"')
            manager.set_last_activated(venv.name)

        if self.option("delete"):
            if not venv:
                self.line_error("error: missing environment name", "error")
                return 1
            if not venv.exists():
                self.line_error(f'error: environment <s>"{venv.name}"</s> does not exist', "error")
                return 1
            venv.delete()
            self.line_error(f'deleted {location} environment <s>"{venv.name}"</s>', "info")

        if self.option("set"):
            assert venv is not None
            if not venv.exists():
                self.line_error(f'error: environment <s>"{venv.name}"</s> does not exist', "error")
                return 1
            manager.set_last_activated(venv.name)
            self.line_error(f'setting environment <s>"{venv.name}"</s> as active', "info")

        if self.option("path"):
            venv = venv or manager.get_last_activated()
            if not venv or not venv.exists():
                if venv and self.argument("name"):
                    self.line_error(f'error: environment <s>"{venv.name}"</s> does not exist', "error")
                else:
                    self.line_error("error: no active environment", "error")
                return 1
            print(venv.path.absolute())

        if self.option("exists"):
            assert venv is not None
            return 0 if venv.exists() else 1

        return 0


class VenvLinkCommand(Command):
    """Link executables from a global virtual environment."""

    name = "venv link"
    arguments = [
        argument(
            "name",
            description="The global environment name.",
        ),
        argument(
            "program",
            description="The name of the program to link.",
        ),
    ]
    options = [
        option(
            "--global",
            "-g",
            description="Manage virtual environments in the global scope instead of the local directory.",
        ),
        option(
            "--force",
            "-f",
            description="Overwrite the link target if it already exists.",
        ),
    ]

    def handle(self) -> int:
        location = "global" if self.option("global") else "local"
        manager = VenvManager(GLOBAL_VENVS_DIRECTORY if self.option("global") else Path(".venvs"))
        venv = manager.get(self.argument("name"))
        if not venv.exists():
            self.line_error(f'error: {location} environment <s>"{venv.name}"</s> does not exist', "error")
            return 1

        program = venv.get_bin(self.argument("program"))
        if not program.is_file():
            self.line_error(
                f'error: program <s>"{program.name}"</s> does not exist in {location} environment <s>"{venv.name}"</s>',
                "error",
            )
            return 1

        target = GLOBAL_BIN_DIRECTORY / program.name
        exists = target.exists() or target.is_symlink()
        if exists and not self.option("force"):
            self.line_error(f'error: target <s>"{target}"</s> already exists', "error")
            return 1

        if exists:
            target.unlink()
        target.symlink_to(program.absolute())
        self.line(f'symlinked <s>"{target}"</s> to <s>"{program}"</s>', "info")

        return 0


class VenvPlugin(ApplicationPlugin):
    def load_configuration(self, app: Application) -> None:
        return None

    def activate(self, app: Application, config: None) -> None:
        app.cleo.add(VenvCommand())
        app.cleo.add(VenvLinkCommand())



import typing as t

from slap.check import Check, CheckResult, check, get_checks
from slap.plugins import CheckPlugin
from slap.project import Project


class GeneralChecksPlugin(CheckPlugin):
    """This plugin provides general checks applicable to all types of projects managed with Slap.

    Plugin ID: `general`."""

    # TODO (@NiklasRosenstein): Check if VCS remote is configured?

    def get_project_checks(self, project: Project) -> t.Iterable[Check]:
        return get_checks(self, project)

    @check("packages")
    def _check_detect_packages(self, project: Project) -> tuple[CheckResult, str | None]:
        """Checks if the project handler employed by Slap for your project is detecting any Python packages. If no
        Python packages can be detected, it might hint at a configuration issue."""

        packages = project.packages()
        result = Check.Result.SKIPPED if packages is None else Check.Result.OK if packages else Check.Result.ERROR
        message = "Detected " + ", ".join(f"<b>{p.root}/{p.name}</b>" for p in packages) if packages else None
        return result, message

    @check("typed")
    def _check_py_typed(self, project: Project) -> tuple[CheckResult, str]:
        expect_typed = project.config().typed
        if expect_typed is None:
            return Check.Result.WARNING, "<b>tool.slap.typed</b> is not set"

        has_py_typed = set[str]()
        has_no_py_typed = set[str]()
        for package in project.packages() or []:
            (has_py_typed if (package.path / "py.typed").is_file() else has_no_py_typed).add(package.name)

        if expect_typed and has_no_py_typed:
            error = True
            message = f'<b>py.typed</b> missing in package(s) <b>{", ".join(has_no_py_typed)}</b>'
        elif not expect_typed and has_py_typed:
            error = True
            message = f'<b>py.typed</b> in package(s) should not exist <b>{", ".join(has_py_typed)}</b>'
        else:
            error = False
            message = (
                "<b>py.typed</b> exists as expected" if expect_typed else "<b>py.typed</b> does not exist as expected"
            )

        return (Check.Result.ERROR if error else Check.Result.OK, message)



from __future__ import annotations

import abc
import typing as t

from nr.util.generic import T

if t.TYPE_CHECKING:
    from pathlib import Path

    from poetry.core.semver.version import Version  # type: ignore[import]

    from slap.application import IO, Application
    from slap.check import Check
    from slap.project import Dependencies, Package, Project
    from slap.python.dependency import Dependency
    from slap.release import VersionRef
    from slap.repository import Repository, RepositoryHost
    from slap.util.vcs import Vcs


class ApplicationPlugin(t.Generic[T], abc.ABC):
    """A plugin that is activated on application load, usually used to register additional CLI commands."""

    ENTRYPOINT = "slap.plugins.application"

    @abc.abstractmethod
    def load_configuration(self, app: Application) -> T:
        """Load the configuration of the plugin. Usually, plugins will want to read the configuration from the Slap
        configuration, which is either loaded from `pyproject.toml` or `slap.toml`. Use #Application.raw_config
        to access the Slap configuration."""

    @abc.abstractmethod
    def activate(self, app: Application, config: T) -> None:
        """Activate the plugin. Register a #Command to #Application.cleo or another type of plugin to
        the #Application.plugins registry."""


class RepositoryHandlerPlugin(abc.ABC):
    """A plugin to provide data and operations on a repository level."""

    ENTRYPOINT = "slap.plugins.repository"

    @abc.abstractmethod
    def matches_repository(self, repository: Repository) -> bool:
        """Return `True` if the handler is able to provide data for the given project."""

    @abc.abstractmethod
    def get_vcs(self, repository: Repository) -> Vcs | None:
        """Return the version control system that the repository is managed with."""

    @abc.abstractmethod
    def get_repository_host(self, repository: Repository) -> RepositoryHost | None:
        """Return the interface for interacting with the VCS hosting service."""

    @abc.abstractmethod
    def get_projects(self, repository: Repository) -> list[Project]:
        """Return the projects of this repository."""


class ProjectHandlerPlugin(abc.ABC):
    """A plugin that implements the core functionality of a project. Project handlers are intermediate layers between
    the Slap tooling and the actual project configuration, allowing different types of configurations to be adapted and
    used with Slap."""

    ENTRYPOINT = "slap.plugins.project"

    @abc.abstractmethod
    def matches_project(self, project: Project) -> bool:
        """Return `True` if the handler is able to provide data for the given project."""

    @abc.abstractmethod
    def get_dist_name(self, project: Project) -> str | None:
        """Return the distribution name for the project."""

    @abc.abstractmethod
    def get_readme(self, project: Project) -> str | None:
        """Return the readme file configured for the project."""

    @abc.abstractmethod
    def get_packages(self, project: Project) -> list[Package] | None:
        """Return a list of packages for the project. Return `None` to indicate that the project is expected to
        not contain any packages."""

    @abc.abstractmethod
    def get_dependencies(self, project: Project) -> Dependencies:
        """Return the dependencies of the project."""

    def get_version(self, project: Project) -> str | None:
        """Return the main project version string."""

        ref = next(iter(self.get_version_refs(project)), None)
        return ref.value if ref else None

    def get_version_refs(self, project: Project) -> list[VersionRef]:
        """Allows the project handler to return additional version refs. Usually returns the version reference in
        `pyproject.toml`."""

        return []

    def add_dependency(self, project: Project, dependency: Dependency, where: str) -> None:
        """Add a dependency to the project configuration.

        Arguments:
          project: The project to update.
          selector: The dependency to add.
          where: The location of where to add the dependency. This is either `'run'`, `'dev'`, or otherwise
            refers to the name of an extra requirement.
          Raises NotImplementedError: If the operation is not supported by the project handler.
        """

        raise NotImplementedError


class CheckPlugin(abc.ABC):
    """This plugin type can be implemented to add custom checks to the `slap check` command. Note that checks will
    be grouped and their names prefixed with the plugin name, so that name does not need to be included in the name
    of the returned checks."""

    ENTRYPOINT = "slap.plugins.check"

    def get_project_checks(self, project: Project) -> t.Iterable[Check]:
        return []

    def get_application_checks(self, app: Application) -> t.Iterable[Check]:
        return []


class ReleasePlugin(abc.ABC):
    """This plugin type provides additional references to the project's version number allowing `slap release` to
    update these references to a new version number.
    """

    ENTRYPOINT = "slap.plugins.release"

    app: Application
    io: IO

    def get_version_refs(self, project: Project) -> list[VersionRef]:
        """Return a list of occurrences of the project version."""

        return []

    def create_release(self, repository: Repository, target_version: str, dry: bool) -> t.Sequence[Path]:
        """Gives the plugin a chance to perform an arbitrary action after all version references have been bumped,
        being informed of the target version. If *dry* is `True`, the plugin should only act as if it was performing
        its usual actions but not commit the changes to disk. It should return the list of files that it modifies
        or would have modified."""

        return []


class VersionIncrementingRulePlugin(abc.ABC):
    """This plugin type can be implemented to provide rules accepted by the `slap release <rule>` command to "bump" an
    existing version number to another. The builtin rules implemented in #slap.ext.version_incrementing_rules.
    """

    ENTRYPOINT = "slap.plugins.version_incrementing_rule"

    def increment_version(self, version: Version) -> Version:
        ...


class ChangelogUpdateAutomationPlugin(abc.ABC):
    """This plugin type can be used with the `slap changelog update-pr -use <plugin_name>` option. It provides all the
    details derivable from the environment (e.g. environment variables available from CI builds) that can be used to
    detect which changelog entries have been added in a pull request, the pull request URL and the means to publish
    the changes back to the original repository.
    """

    ENTRYPOINT = "slap.plugins.changelog_update_automation"

    io: IO

    @abc.abstractmethod
    def initialize(self) -> None:
        ...

    @abc.abstractmethod
    def get_base_ref(self) -> str:
        ...

    @abc.abstractmethod
    def get_pr(self) -> str:
        ...

    @abc.abstractmethod
    def publish_changes(self, changed_files: list[Path]) -> None:
        ...



import logging
import os
import typing as t
from pathlib import Path

from nr.util.singleton import NotSet

from slap.application import IO, Application, argument, option
from slap.ext.application.venv import VenvAwareCommand
from slap.plugins import ApplicationPlugin
from slap.project import Project

logger = logging.getLogger(__name__)


class TestRunner:

    _colors = ["blue", "cyan", "magenta", "yellow"]
    _prev_color: t.ClassVar[str | None] = None

    def __init__(self, name: str, config: t.Any, io: IO, cwd: Path | None = None, line_prefixing: bool = True) -> None:
        assert isinstance(config, str), type(config)
        self.name = name
        self.config = config
        self.io = io
        self.cwd = cwd
        self.line_prefixing = line_prefixing

    def run(self) -> int:
        import subprocess as sp
        import sys
        from codecs import getreader

        from cleo.io.io import OutputType  # type: ignore[import]
        from ptyprocess import PtyProcessUnicode  # type: ignore[import]

        color = (
            self._colors[0]
            if TestRunner._prev_color is None
            else self._colors[(self._colors.index(TestRunner._prev_color) + 1) % len(self._colors)]
        )
        TestRunner._prev_color = color
        command = ["bash", "-c", self.config]
        prefix = f"{self.name}| "

        logger.info("Running command <subj>%s</subj> in <val>%s</val>", command, self.cwd)

        try:
            cols, rows = os.get_terminal_size()
        except OSError:
            sproc = sp.Popen(command, cwd=self.cwd, stdout=sp.PIPE, stderr=sp.STDOUT)
            assert sproc.stdout
            stdout = getreader(sys.getdefaultencoding())(sproc.stdout)
            for line in iter(stdout.readline, ""):
                line = line.rstrip()
                if self.line_prefixing:
                    self.io.write(f"<fg={color}>{prefix}</fg>")
                self.io.write(line + "\n", type=OutputType.NORMAL)
            sproc.wait()
            assert sproc.returncode is not None
            return sproc.returncode
        else:
            proc = PtyProcessUnicode.spawn(command, dimensions=(rows, cols - len(prefix)), cwd=self.cwd)
            while not proc.eof():
                try:
                    line = proc.readline().rstrip()
                except EOFError:
                    break
                if self.line_prefixing:
                    self.io.write(f"<fg={color}>{prefix}</fg>")
                self.io.write(line + "\n", type=OutputType.NORMAL)
            proc.wait()
            assert proc.exitstatus is not None
            return proc.exitstatus


class Test(t.NamedTuple):
    project: Project
    name: str
    command: str

    @property
    def id(self) -> str:
        return f"{self.project.id}:{self.name}"


class TestCommandPlugin(VenvAwareCommand, ApplicationPlugin):
    """
    Execute commands configured in <code>[tool.slap.test]</code>.

    <b>Example configuration:</b>

      <fg=cyan>[tool.slap.test]</fg>
      <fg=green>pytest</fg> = <fg=yellow>"pytest --cov=my_package tests/"</fg>
      <fg=green>mypy</fg> = <fg=yellow>"mypy src"</fg>

    <b>Example usage:</b>

      <fg=yellow>$</fg> slap test
      <fg=dark_gray>mypy | Success: no issues found in 12 source files
      pytest | ===================================== test session starts ======================================
      pytest | platform linux -- Python 3.10.2, pytest-6.2.5, py-1.11.0, pluggy-1.0.0
      ...</fg>
    """

    app: Application
    tests: list[Test]
    requires_venv: t.ClassVar[bool] = False

    name = "test"
    arguments = [
        argument("test", "One or more tests to run (runs all if none are specified)", optional=True, multiple=True),
    ]
    options = VenvAwareCommand.options + [
        option(
            "--no-line-prefix",
            "-s",
            description="Do not prefix output from the test commands with the test name (default if a single argument "
            "for <info>test</info> is specified).",
        ),
        option(
            "--list",
            "-l",
            description="List all available tests",
        ),
        option(
            "--exclude",
            "-x",
            description="Do not run the specified test. Can be passed multiple times.",
            flag=False,
            multiple=True,
        ),
    ]

    # Hack to set a default value for the flag
    next(opt for opt in options if opt.name == "no-line-prefix")._default = NotSet.Value

    def load_configuration(self, app: Application) -> None:
        self.app = app
        self.tests = []
        for project in app.repository.projects():
            for test_name, command in project.raw_config().get("test", {}).items():
                self.tests.append(Test(project, test_name, command))

    def activate(self, app: Application, config: None) -> None:
        app.cleo.add(self)

    def _select_tests(self, name: str) -> set[Test]:
        result = set()
        for test in self.tests:
            use_test = (
                self.app.repository.is_monorepo
                and (name == test.id or (name.startswith(":") and test.name == name[1:]) or (test.project.id == name))
                or not self.app.repository.is_monorepo
                and (name == test.name)
            )
            if use_test:
                result.add(test)
        if not result:
            raise ValueError(f"{name!r} did not match any tests")
        return result

    def handle(self) -> int:
        result = super().handle()
        if result != 0:
            return result
        if self.option("list"):
            if self.argument("test"):
                self.line_error("error: incompatible arguments (<opt>test</opt> and <opt>-l,--list</opt>)", "error")
                return 1
            for test in self.tests:
                print(test.id)
            return 0

        if not self.tests:
            self.line_error("error: no tests configured", "error")
            return 1

        test_names: list[str] = self.argument("test")
        exclude_tests: list[str] = self.option("exclude")

        if not test_names:
            tests = set(self.tests)
        else:
            try:
                tests = {t for a in test_names for t in self._select_tests(a)}
            except ValueError as exc:
                self.line_error(f"error: {exc}", "error")
                return 1

        tests -= {t for a in exclude_tests for t in self._select_tests(a)}

        if (no_line_prefix := self.option("no-line-prefix")) is NotSet.Value:
            no_line_prefix = test_names is not None and len(tests) == 1

        single_project = len(set(t.project for t in self.tests)) == 1

        results = {}
        for test in sorted(tests, key=lambda t: t.id):
            results[test.name if single_project else test.id] = TestRunner(
                test.name if single_project else test.id,
                test.command,
                self.io,
                test.project.directory,
                not no_line_prefix,
            ).run()

        if len(tests) > 1:
            self.line("\n<comment>test summary:</comment>")
            for test_name, exit_code in results.items():
                color = "green" if exit_code == 0 else "red"
                self.line(f"  <fg={color}>•</fg> {test_name} (exit code: {exit_code})")

        return 0 if set(results.values()) == {0} else 1



from __future__ import annotations

import abc
import dataclasses
import typing as t
from pathlib import Path

from databind.core.settings import Union
from nr.util.functional import Once

from slap.configuration import Configuration

if t.TYPE_CHECKING:
    from slap.plugins import RepositoryHandlerPlugin
    from slap.project import Project
    from slap.util.vcs import Vcs


@dataclasses.dataclass
class Issue:
    """Represents an issue."""

    id: str
    url: str
    shortform: str


@dataclasses.dataclass
class PullRequest:
    """Represents a pull request."""

    id: str
    url: str
    shortform: str


@Union("!slap.plugins.repository_host")
class RepositoryHost(abc.ABC):
    """Interface for repository hosting services to resolve issue and pull request references, comment on issues
    and create releases."""

    ENTRYPOINT = "slap.plugins.repository_host"

    @abc.abstractmethod
    def get_username(self, repository: Repository) -> str | None:
        ...

    @abc.abstractmethod
    def get_issue_by_reference(self, issue_reference: str) -> Issue:
        ...

    @abc.abstractmethod
    def get_pull_request_by_reference(self, pr_reference: str) -> PullRequest:
        ...

    @abc.abstractmethod
    def comment_on_issue(self, issue_reference: str, message: str) -> None:
        ...

    @abc.abstractmethod
    def create_release(self, version: str, description: str, attachments: list[Path]) -> None:
        ...

    @staticmethod
    @abc.abstractmethod
    def detect_repository_host(repository: Repository) -> RepositoryHost | None:
        ...


class Repository(Configuration):
    """A repository represents a directory that contains one or more projects. A repository represents one or more
    projects in one logical unit, usually tracked by a single version control repository. The class"""

    handler: Once[RepositoryHandlerPlugin]

    def __init__(self, directory: Path) -> None:
        super().__init__(directory)
        self._handler = Once(self._get_repository_handler)
        self.projects = Once(self._get_projects)
        self.vcs = Once(self._get_vcs)
        self.host = Once(self._get_repository_host)

    @property
    def is_monorepo(self) -> bool:
        if len(self.projects()) > 1 or (len(self.projects()) == 1 and self.projects()[0].directory != self.directory):
            return True
        return False

    def _get_repository_handler(self) -> RepositoryHandlerPlugin | None:
        """Returns the handler for this repository."""

        from nr.util.plugins import load_entrypoint

        from slap.ext.repository_handlers.default import DefaultRepositoryHandler
        from slap.plugins import RepositoryHandlerPlugin

        handler: RepositoryHandlerPlugin
        handler_name = self.raw_config().get("repository", {}).get("handler")
        if handler_name is None:
            handler = DefaultRepositoryHandler()
            if not handler.matches_repository(self):
                return None
        else:
            assert isinstance(handler_name, str), repr(handler_name)
            handler = load_entrypoint(RepositoryHandlerPlugin, handler_name)()  # type: ignore[misc]
            assert handler.matches_repository(self), (self, handler)
        return handler

    def _get_projects(self) -> list[Project]:
        """Returns the projects provided by the project handler, but sorted in topological order."""

        handler = self._handler()
        if not handler:
            return []

        return sorted(handler.get_projects(self), key=lambda p: p.id)

    def get_projects_ordered(self) -> list[Project]:
        """Return a topological ordering of the projects."""

        from nr.util.digraph import DiGraph
        from nr.util.digraph.algorithm.topological_sort import topological_sort

        from slap.project import Project

        graph: DiGraph[Project, None, None] = DiGraph()
        projects = self.projects()
        for project in projects:
            assert isinstance(project, Project)
            graph.add_node(project, None)
            for dep in project.get_interdependencies(projects):
                graph.add_node(dep, None)
                graph.add_edge(dep, project, None)

        return list(topological_sort(graph, sorting_key=lambda p: p.id))

    def _get_vcs(self) -> Vcs | None:
        from nr.util.optional import Optional

        return Optional(self._handler()).map(lambda h: h.get_vcs(self)).or_else(None)

    def _get_repository_host(self) -> RepositoryHost | None:
        from nr.util.optional import Optional

        return Optional(self._handler()).map(lambda h: h.get_repository_host(self)).or_else(None)



import collections
import dataclasses
import logging
import typing as t

from nr.util.plugins import load_entrypoint

from slap.application import Application, Command, option
from slap.check import Check, CheckResult
from slap.plugins import ApplicationPlugin, CheckPlugin
from slap.project import Project

logger = logging.getLogger(__name__)
DEFAULT_PLUGINS = ["changelog", "general", "poetry", "release"]
COLORS = {
    Check.Result.OK: "green",
    Check.Result.RECOMMENDATION: "magenta",
    Check.Result.WARNING: "yellow",
    Check.Result.ERROR: "red",
    Check.Result.SKIPPED: "light_gray",
}


@dataclasses.dataclass
class CheckConfig:
    #: A list of checks that are enabled for the project.
    plugins: list[str] = dataclasses.field(default_factory=lambda: DEFAULT_PLUGINS[:])


class CheckCommandPlugin(Command, ApplicationPlugin):
    """Run sanity checks on your Python project."""

    app: Application
    config: dict[Project, CheckConfig]

    name = "check"
    options = [
        option(
            "--show-skipped",
            description="Show skipped checks.",
        ),
        option(
            "--warnings-as-errors",
            "w",
            description="Treat warnings as errors.",
        ),
    ]

    def load_configuration(self, app: "Application") -> dict[Project, CheckConfig]:
        import databind.json

        result = {}
        for project in app.repository.projects():
            config = databind.json.load(project.raw_config().get("check", {}), CheckConfig)
            result[project] = config
        return result

    def activate(self, app: "Application", config: dict[Project, CheckConfig]) -> None:
        self.app = app
        self.config = config
        app.cleo.add(self)

    def handle(self) -> int:

        counter: t.MutableMapping[CheckResult, int] = collections.defaultdict(int)
        if self.app.repository.is_monorepo:
            for check in self._run_application_checks():
                counter[check.result] += 1
        for project in self.app.repository.projects():
            if not project.is_python_project:
                continue
            for check in self._run_project_checks(project):
                counter[check.result] += 1

        if self.option("warnings-as-errors") and counter.get(Check.WARNING, 0) > 0:
            exit_code = 1
        elif counter.get(Check.ERROR, 0) > 0:
            exit_code = 1
        else:
            exit_code = 0

        self.line(
            "Summary: "
            + ", ".join(
                f"{count} <fg={COLORS[result]};options=bold>{result.name}</fg>"
                for result, count in sorted(counter.items())
            )
            + f", exit code: {exit_code}"
        )

        return exit_code

    def _print_checks(self, checks: t.Sequence[Check]) -> None:
        max_w = max(len(c.name) for c in checks)
        for check in checks:

            if not self.option("show-skipped") and check.result == Check.SKIPPED:
                continue

            color = COLORS[check.result]
            self.io.write(
                f"  <b>{check.name.ljust(max_w)}</b>  <fg={color};options=bold>{check.result.name.ljust(14)}</fg>"
            )
            if check.description:
                self.io.write(f" — {check.description}")
            self.io.write("\n")

            if check.details:
                for line in check.details.splitlines():
                    self.io.write_line(f"    {line}")

    def _run_project_checks(self, project: Project) -> t.Iterator[Check]:
        checks = []
        for plugin_name in sorted(self.config[project].plugins):
            plugin = load_entrypoint(CheckPlugin, plugin_name)()
            try:
                for check in sorted(plugin.get_project_checks(project), key=lambda c: c.name):
                    check.name = f"{plugin_name}:{check.name}"
                    yield check
                    checks.append(check)
            except Exception as exc:
                logger.exception(
                    "Uncaught exception in project <subj>%s</subj> application checks for plugin <val>%s</val>",
                    project,
                    plugin_name,
                )
                check = Check(f"{plugin_name}", CheckResult.ERROR, str(exc))
                yield check
                checks.append(check)
            if not self.app.repository.is_monorepo:
                try:
                    for check in sorted(plugin.get_application_checks(self.app), key=lambda c: c.name):
                        check.name = f"{plugin_name}:{check.name}"
                        yield check
                        checks.append(check)
                except Exception as exc:
                    logger.exception("Uncaught exception in project checks for plugin <val>%s</val>", plugin_name)
                    check = Check(f"{plugin_name}", CheckResult.ERROR, str(exc))
                    yield check
                    checks.append(check)

        if checks:
            if self.app.repository.is_monorepo:
                self.line(f"Checks for project <info>{project.id}</info>")
                self.line("")
            self._print_checks(checks)
            self.line("")

    def _run_application_checks(self) -> t.Iterable[Check]:
        plugin_names = {p for project in self.app.repository.projects() for p in self.config[project].plugins}
        checks = []
        for plugin_name in sorted(plugin_names):
            plugin = load_entrypoint(CheckPlugin, plugin_name)()
            try:
                for check in sorted(plugin.get_application_checks(self.app), key=lambda c: c.name):
                    check.name = f"{plugin_name}:{check.name}"
                    yield check
                    checks.append(check)
            except Exception as exc:
                logger.exception("Uncaught exception in application checks for plugin <val>%s</val>", plugin_name)
                check = Check(f"{plugin_name}", CheckResult.ERROR, str(exc))
                yield check
                checks.append(check)

        if checks:
            self.line("Global checks:")
            self._print_checks(checks)
            self.line("")



from __future__ import annotations

import dataclasses
import sys
import typing as t
from pathlib import Path

from databind.core.settings import Alias, ExtraKeys

from slap.application import Application, Command, argument, option
from slap.configuration import Configuration
from slap.plugins import ApplicationPlugin, ReleasePlugin, VersionIncrementingRulePlugin

if t.TYPE_CHECKING:
    from poetry.core.semver.version import Version  # type: ignore[import]

    from slap.project import Project
    from slap.release import VersionRef


@dataclasses.dataclass
class VersionRefConfig:
    """Version reference."""

    #: The file in which the reference occurs.
    file: str

    #: A regex pattern that matches the version. Must not contain any captured groups except for one that represents
    #: the version number itself. The expression for that group can be writen as `{version}` instead (example:
    #: `version: "{version}"`).
    pattern: str


@dataclasses.dataclass
@ExtraKeys(True)
class ReleaseConfig:
    #: The VCS branch on which releases are allowed. The release command will prevent you from creating a release
    #: while on a different branch (unless `--no-branch-check` is set).
    branch: str = "develop"

    #: The template for the commit message when a release is created and the `--tag, -t` option is used.
    commit_message: t.Annotated[str, Alias("commit-message")] = "release {version}"

    #: The template for the tag name when a release is created and the `--tag, -t` option is used.
    tag_format: t.Annotated[str, Alias("tag-format")] = "{version}"

    #: A list of references to the version number that should be updated along with the version numbers
    #: that the release command knows about by default (like the `version` in `pyproject.toml` and the
    #: version number in the source code).
    references: list[VersionRefConfig] = dataclasses.field(default_factory=list)

    #: A list of #ReleasePlugins to use. Defaults to contain the #SourceCodeVersionReferencesPlugin
    #: and #ChangelogReleasePlugin.
    plugins: list[str] = dataclasses.field(default_factory=lambda: ["changelog_release", "source_code_version"])


class ReleaseCommandPlugin(Command, ApplicationPlugin):
    """Create a new release of your Python package.

    This command can perform the following operations in sequence (most of them need
    to be enabled explicitly with flags):

    1. Bump the version number in <u>pyproject.toml</u> and in other files
    2. Add the changed files to Git, create a commit and a tag (<opt>--tag, -t</opt>)
    3. Push the commit and tag to the remote repository (<opt>--push, -p</opt>)

    In addition to the <u>pyproject.toml</u>, the command will automatically detect the file(s)
    in your Python source code that contain a <b>__version__</b> member and update it as well.
    Additional files can be updated by configuring the <fg=green>[tool.slap.release.references]</fg>
    option:

      <fg=green>[tool.slap.release]</fg>
      <fg=dark_gray>references</fg> = [
        {{ <fg=dark_gray>file</fg> = <fg=yellow>"../frontend/package.json"</fg>, <fg=dark_gray>pattern</fg> = <fg=yellow>"  \\"version\\": \\"{{VERSION}}\\""</fg> }}
      ]

    Furthermore, the <opt>--validate</opt> option can be used in CI to ensure that the version
    numbers are consistent across the project. This is particularly useful when
    automating publishing from CI builds.

    <b>Bumping the version number</b>

      Specifying an explicit version number or a version bump rule for the <opt>version</opt>
      argument will update the version across all references that can be detected.
      You can use <opt>--validate</opt> to show all files in which version numbers are found.

      The supported rules are <u>major</u>, <u>premajor</u>, <u>minor</u>, <u>preminor</u>, <u>patch</u>, <u>prepatch</u> and
      <u>prerelease</u>.

      <u>[Git]</u>: The command will prevent you from bumping the version unless you are on
      the branch configured under <fg=green>[tool.slap.release.branch]</fg> or "develop" by default.
      If you want to skip that check, pass <opt>--no-branch-check</opt>.

    <b>Commit & tag</b>

      <u>[Git]</u>: You can use the <opt>--tag, -t</opt> flag to automatically add the updated files,
      create a new commit and tag the commit with the version number. The tag name
      by default will just be the version number, but can be changed by setting the
      <fg=green>[tool.slap.release.tag_format]</fg>. Similarly, the commit message used can be
      configured with <fg=green>[tool.slap.release.commit_message]</fg>.

    <b>Push to remote</b>

      <u>[Git]</u>: Using the <opt>--push, -p</opt> in combination with <opt>--tag, -t</opt> will push the new
      commit and tag to the remote Git repository immediately. You can specify the
      <opt>--remote, -r</opt> option to change the remote which will be pushed to (defaults
      to "origin").
    """  # noqa: E501

    app: Application
    config: dict[Configuration, ReleaseConfig]

    name = "release"
    arguments = [
        argument("version", "The target version number or rule to apply to the current version.", True),
    ]
    options = [
        option("tag", "t", "Create a Git tag after the version numbers were updated."),
        option("push", "p", "Push the changes to the Git remote repository."),
        option("remote", "r", "The Git remote to push to (only when <opt>--push</opt> is specified).", False),
        option("dry", "d", "Do not commit changes to disk."),
        option("force", "f", "Force tag creation and push."),
        option(
            "validate",
            None,
            "Instead of bumping the version, validate that all version references are consistent.\n"
            "If the <opt>version</opt> argument is specified, all version references must match it.",
        ),
        option(
            "no-branch-check", None, "Do not validate the current Git branch matches the configured release branch."
        ),
        option("no-worktree-check", None, "Do not check the worktree state."),
    ]

    # TODO (@NiklasRosenstein): Support "git" rule for bumping versions

    def load_configuration(self, app: Application) -> dict[Configuration, ReleaseConfig]:
        import databind.json

        result = {}
        for project in t.cast(list[Configuration], [app.repository] + app.repository.projects()):  # type: ignore[operator]  # noqa: E501
            data = project.raw_config().get("release", {})
            result[project] = databind.json.load(data, ReleaseConfig)
        self.app = app
        self.config = result
        return result

    def activate(self, app: Application, config: dict[Project, ReleaseConfig]) -> None:
        app.cleo.add(self)

    def _validate_options(self) -> int:
        """Internal. Ensures that the combination of provided options make sense."""

        if self.option("dry") and self.option("validate"):
            self.line_error("error: <opt>--dry</opt> cannot be combined with <opt>--validate</opt>", "error")
            return 1
        if self.option("tag") and self.option("validate"):
            self.line_error("error: <opt>--tag</opt> cannot be combined with <opt>--validate</opt>", "error")
            return 1
        if self.option("push") and not self.option("tag"):
            self.line_error("error: <opt>--push</opt> can only be combined with <opt>--tag</opt>", "error")
            return 1
        if self.option("force") and not self.option("tag"):
            self.line_error(
                "error: <opt>--force</opt> can only be combined with <opt>--tag</opt> and <opt>--push</opt>", "error"
            )
            return 1
        if self.option("remote") is not None and not self.option("push"):
            self.line_error("error: <opt>--remote</opt> can only be combined with <opt>--push</opt>", "error")
            return 1

        self.io.input.set_option("remote", self.option("remote") or "origin")

        if self.option("tag") and not self.is_git_repository:
            self.line_error("error: not in a git repository, cannot use <opt>--tag</opt>", "error")
            return 1
        if self.option("push") and not self.is_git_repository:
            self.line_error("error: not in a git repository, cannot use <opt>--push</opt>", "error")
            return 1
        if self.option("push") and (remote := self.option("remote")) not in {r.name for r in self.git.remotes()}:
            self.line_error(f'error: git remote "{remote}" does not exist', "error")
            return 1

        return 0

    def _load_plugins(self, configuration: Configuration) -> list[ReleasePlugin]:
        """Internal. Loads the plugins for the given configuration."""

        from nr.util.plugins import load_entrypoint

        plugins = []
        for plugin_name in self.config[configuration].plugins:
            plugin = load_entrypoint(ReleasePlugin, plugin_name)()
            plugin.app = self.app
            plugin.io = self.io
            plugins.append(plugin)
        return plugins

    def _show_version_refs(self, version_refs: list[VersionRef], increment_to: str | None = None) -> None:
        """Internal. Prints the version references to the terminal."""

        max_w1 = max(len(str(ref.file)) for ref in version_refs) + 1
        max_w2 = max(len(ref.value) for ref in version_refs)
        prev: VersionRef | None = None
        for ref in sorted(version_refs, key=lambda r: r.file):
            filename = str(ref.file) + ":" if not prev or prev.file != ref.file else ""
            self.io.write(f"  <fg=cyan>{(filename).ljust(max_w1)}</fg> {ref.value.ljust(max_w2)}")
            if increment_to:
                self.io.write(f" → <b>{increment_to}</b>")
            self.io.write_line(f" <fg=dark_gray># {ref.content!r}</fg>")
            prev = ref

    def _validate_version_refs(self, version_refs: list[VersionRef], version: str | None) -> int:
        """Internal. Verifies the consistency of the given version references. This is used when `--validate` is set."""

        from poetry.core.semver.version import Version

        if version is not None:
            Version.parse(version)

        versions = set(ref.value for ref in version_refs)
        if not versions:
            self.line("<info>no version numbers detected</info>")
            return 1

        if len(versions) > 1:
            self.line("<error>versions are inconsistent</error>")
            self._show_version_refs(version_refs)
            return 1

        has_version = next(iter(versions))
        if version is not None:
            if version != has_version:
                self.line(f"<error>version mismatch, expected <b>{version}</b>, got <b>{has_version}</b></error>")
                return 1

        self.line("<comment>versions are ok</comment>")
        self._show_version_refs(version_refs)
        return 0

    def _check_on_release_branch(self) -> bool:
        """Internal. Checks if the current Git branch matches the configured release branch."""

        from nr.util.git import NoCurrentBranchError

        if not self.is_git_repository or self.option("no-branch-check"):
            return True

        config = self.config[self.app.repository]

        try:
            current_branch = self.git.get_current_branch_name()
        except NoCurrentBranchError:
            self.line_error("error: not currently on a Git branch", "error")
            return False

        if current_branch != config.branch:
            self.line_error(
                f"error: current branch is <b>{current_branch}</b> but must be on the "
                f"release branch (<b>{config.branch}</b>)",
                "error",
            )
            return False
        return True

    def _check_clean_worktree(self, required_files: list[Path]) -> bool:
        """Internal. Checks that the Git work state is clean and that all the *required_files* are tracked
        in the repo."""

        if not self.is_git_repository or self.option("no-worktree-check"):
            return True

        queried_files = {f.resolve() for f in required_files}
        tracked_files = {Path(f).resolve() for f in self.git.get_files()}
        if untracked_files := queried_files - tracked_files:
            self.line_error("error: some of the files with version references are not tracked by Git", "error")
            for fn in untracked_files:
                self.line_error(f"  · {fn}", "error")
            return False

        file_status = list(self.git.get_status())
        if any(f.mode[1] != " " for f in file_status):
            self.line_error("error: found untracked changes in worktree", "error")
            return False
        if any(f.mode[0] not in " ?" for f in file_status):
            self.line(
                "<fg=yellow>found modified files in the staging area. these files will be committed into "
                "the release tag.</fg>"
            )
            if not self.confirm("continue?"):
                return False

        return True

    def _get_current_version(self, version_refs: list[VersionRef]) -> "Version":
        """Try to identify the current version number among the version refs. This is done by selecting all versions
        that occur in a `pyproject.toml`, and if they are all equal, they are considered the current version. If they
        are different, a #ValueError is raised."""

        from poetry.core.semver.version import Version

        current_version = {r.value for r in version_refs}
        if len(current_version) != 1:
            raise ValueError("could not determine current version number")

        return Version.parse(next(iter(current_version)))

    def _get_new_version(self, version_refs: list[VersionRef], rule: str) -> "Version":
        """Return the new version, based on *rule*. If *rule* is a version string, it is used as the new version.
        Otherwise, it is considered a rule and the applicable rule plugin is invoked to construct the new version."""

        from nr.util.plugins import NoSuchEntrypointError, load_entrypoint
        from poetry.core.semver.version import Version

        try:
            return Version.parse(rule)
        except ValueError:
            try:
                plugin = load_entrypoint(VersionIncrementingRulePlugin, rule)
            except NoSuchEntrypointError:
                self.line(f'error: "<b>{rule}</b>" is not a valid version incrementing rule', "error")
                sys.exit(1)
            return plugin().increment_version(self._get_current_version(version_refs))

    def _bump_version(self, version_refs: list[VersionRef], target_version: Version, dry: bool) -> list[Path]:
        """Internal. Replaces the version reference in all files with the specified *version*."""

        from nr.util import Stream
        from nr.util.text import substitute_ranges

        self.line(
            f'bumping <b>{len(version_refs)}</b> version reference{"" if len(version_refs) == 1 else "s"} to '
            f"<b>{target_version}</b>"
        )

        changed_files: list[Path] = []

        self._show_version_refs(version_refs, str(target_version))
        self.line("")
        for filename, refs in Stream(version_refs).groupby(lambda r: r.file):
            with open(filename) as fp:
                content = fp.read()

            content = substitute_ranges(
                content,
                ((ref.start, ref.end, str(target_version)) for ref in refs),
            )

            changed_files.append(filename)
            if not dry:
                with open(filename, "w") as fp:
                    fp.write(content)

        for plugin in self._load_plugins(self.app.repository):
            try:
                changed_files.extend(plugin.create_release(self.app.repository, str(target_version), dry))
            except BaseException:
                self.line_error(f"error with {type(plugin).__name__}.bump_version()", "error")
                raise

        return changed_files

    def _create_tag(self, target_version: str, changed_files: list[Path], dry: bool, force: bool) -> str:
        """Internal. Used when --tag is specified to create a Git tag."""

        assert self.is_git_repository

        # TODO (@NiklasRosenstein): If this step errors, revert the changes made by the command so far?

        config = self.config[self.app.repository]

        if "{version}" not in config.tag_format:
            self.line_error("<info>tool.slap.release.tag-format<info> must contain <info>{version}</info>", "error")
            sys.exit(1)
        tag_name = config.tag_format.replace("{version}", str(target_version))
        self.line("")
        self.line(f"tagging <fg=cyan>{tag_name}</fg>")

        if not dry:
            commit_message = config.commit_message.replace("{version}", str(target_version))
            self.git.add([str(f) for f in changed_files])
            self.git.commit(commit_message, allow_empty=True)
            self.git.tag(tag_name, force=force)

        return tag_name

    def _push_to_remote(self, tag_name: str, remote: str, dry: bool, force: bool) -> None:
        """Internal. Push the current branch and the tag to the remote repository. Use when `--push` is set."""

        assert self.is_git_repository

        branch = self.git.get_current_branch_name()

        self.line("")
        self.line(f"pushing <fg=cyan>{branch}</fg>, <fg=cyan>{tag_name}</fg> to <info>{remote}</info>")

        if not dry:
            self.git.push(remote, branch, tag_name, force=force)

    def _get_version_refs(self) -> list[VersionRef]:
        """Extracts all version references in the projects controlled by the application and returns them."""

        from slap.project import Project
        from slap.release import match_version_ref_pattern

        version_refs = []

        # Understand the version references defined in the project configuration.
        for project in self.app.configurations():

            # Always consider the version number in the pyproject.toml.
            if project.pyproject_toml.exists() and isinstance(project, Project):
                version_refs += project.get_version_refs()

            for config in self.config[project].references:
                pattern = config.pattern.replace("{version}", r"(.*?)")
                version_ref = match_version_ref_pattern(project.directory / config.file, pattern)
                if version_ref and version_ref.value == "":
                    self.line_error(
                        f"warning: custom reference <b>{config}</b> matches an empty string.\n  make sure you have "
                        f"at least one character after the <b>{{version}}</b> construct "
                        "(ex. <b>version: {{version}}$</b>)",
                        "warning",
                    )
                if version_ref is not None:
                    version_refs.append(version_ref)

        # Query plugins for additional version references.
        for project in self.app.repository.projects():
            for plugin in self._load_plugins(project):
                version_refs += plugin.get_version_refs(project)

        version_refs.sort(key=lambda r: r.file)

        for ref in version_refs:
            if ref.file == ref.file.absolute():
                ref.file = ref.file.relative_to(Path.cwd())

        return version_refs

    def handle(self) -> int:
        """Entrypoint for the command."""

        from nr.util.git import Git

        self.git = Git()
        self.is_git_repository = self.git.get_toplevel() is not None

        if (err := self._validate_options()) != 0:
            return err

        version_refs = self._get_version_refs()
        version = self.argument("version")

        if self.option("validate"):
            if version is None:
                try:
                    version = str(self._get_current_version(version_refs))
                except ValueError:
                    pass
            return self._validate_version_refs(version_refs, version)

        if version is not None:
            if self.option("tag") and not self._check_on_release_branch():
                return 1
            if self.option("tag") and not self._check_clean_worktree([x.file for x in version_refs]):
                return 1
            if self.option("dry"):
                self.line_error("dry mode enabled, no changes will be committed to disk", "comment")
            target_version = self._get_new_version(version_refs, version)
            changed_files = self._bump_version(version_refs, target_version, self.option("dry"))
            if self.option("tag"):
                tag_name = self._create_tag(
                    str(target_version), changed_files, self.option("dry"), self.option("force")
                )
                if self.option("push"):
                    self._push_to_remote(tag_name, self.option("remote"), self.option("dry"), self.option("force"))

        else:
            self.line_error(
                "<error>error: no action implied, specify a <info>version</info> argument or the "
                "<info>--validate</info> option"
            )
            return 1

        return 0



""" Project handler for projects using the Setuptools build system. """

from __future__ import annotations

import re
import typing as t

from slap.ext.project_handlers.base import BaseProjectHandler, interdependencies_enabled
from slap.project import Dependencies, Package, Project
from slap.release import VersionRef, match_version_ref_pattern, match_version_ref_pattern_on_lines

if t.TYPE_CHECKING:
    from slap.python.dependency import VersionSpec


class SetuptoolsProjectHandler(BaseProjectHandler):
    def __init__(self) -> None:
        self._project: Project | None = None
        self._setup_cfg: t.Dict[str, t.Dict[str, t.Any]] | None = None

    def _get_setup_cfg(self, project: Project) -> t.Dict[str, t.Any]:
        import configparser

        if self._project is None:
            parser = configparser.SafeConfigParser()
            parser.read(project.directory / "setup.cfg")
            self._setup_cfg = {s: dict(parser.items(s)) for s in parser.sections()}
            self._project = project
        else:
            assert self._project is project
            assert self._setup_cfg is not None
        return self._setup_cfg

    # ProjectHandlerPlugin

    def matches_project(self, project: Project) -> bool:
        if not project.pyproject_toml.exists():
            return False
        build_backend = project.pyproject_toml.get("build-system", {}).get("build-backend")
        return build_backend == "setuptools.build_meta"

    def get_dist_name(self, project: Project) -> str | None:
        return self._get_setup_cfg(project).get("metadata", {}).get("name")

    def get_readme(self, project: Project) -> str | None:
        cfg = self._get_setup_cfg(project)
        long_description = cfg.get("metadata", {}).get("long_description", "").strip()
        if long_description.startswith("file:"):
            return (
                long_description.partition("file:")[2].partition(",")[0].strip()
            )  # Apparently can be a comma separated list of multiple files
        return None

    def get_packages(self, project: Project) -> list[Package] | None:
        # TODO (@NiklasRosenstein): Handle namespace_packages as well
        options = self._get_setup_cfg(project).get("options", {})
        packages: str | None = options.get("packages")
        if packages is None or packages.strip() == "find:":
            # Auto detect packages
            # TODO (@NiklasRosenstein): Limit to the package_dir in setup.cfg?
            return super().get_packages(project)
        elif packages.strip() == "":
            return None  # Intentionally no packages
        package_dir = project.directory / options.get("package_dir", ".")
        return [Package(p, package_dir / p, package_dir) for p in parse_list_semi(packages)]

    def get_dependencies(self, project: Project) -> Dependencies:
        from slap.python.dependency import PypiDependency, VersionSpec

        options = self._get_setup_cfg(project).get("options", {})
        return Dependencies(
            VersionSpec(options["python_requires"]) if "python_requires" in options else None,
            PypiDependency.parse_list(parse_list_semi(options.get("install_requires", ""))),
            PypiDependency.parse_list(
                parse_list_semi(options.get("setup_requires", "")) + parse_list_semi(options.get("tests_require", ""))
            ),
            {
                extra: PypiDependency.parse_list(parse_list_semi(value))
                for extra, value in options.get("extras_require", {}).items()
            },
            PypiDependency.parse_list(parse_list_semi(options.get("setup_requires", ""))),
        )

    def get_version_refs(self, project: Project) -> list[VersionRef]:
        """Returns the version reference in `setup.cfg`."""

        VERSION_PATTERN = r"^version\s*=\s*(.*?)$"
        version_ref = match_version_ref_pattern(project.directory / "setup.cfg", VERSION_PATTERN, None)
        refs = [version_ref] if version_ref else []
        if interdependencies_enabled(project):
            refs += get_setup_cfg_interdependency_version_refs(project)
        return refs

    def get_add_dependency_toml_location_and_config(
        self,
        project: Project,
        package: str,
        version_spec: VersionSpec,
        where: str,
    ) -> tuple[list[str], list | dict]:
        raise NotImplementedError


def parse_list_semi(val: str) -> list[str]:
    """Parses a string to a list of strinsg according to the `list-semi` specification in the [setuptools docs][1].

    [1]: https://setuptools.pypa.io/en/latest/userguide/declarative_config.html#specifying-values"""

    from nr.util.stream import Stream

    return Stream(val.splitlines()).map(lambda v: v.split(";")).concat().map(str.strip).filter(bool).collect()


def get_setup_cfg_interdependency_version_refs(project: Project) -> list[VersionRef]:
    setup_cfg = project.directory / "setup.cfg"
    other_projects: list[str] = [
        t.cast(str, p.dist_name())
        for p in project.repository.projects()
        if p.is_python_project and p is not project and p.dist_name()
    ]

    refs = []
    for project_name in other_projects:
        # Look for occurrences of the project name in the context of requirements.
        expressions = [
            # Match requirements split over multiple lines.
            r"^\w+_requires?\s*=.*^\s+(?:" + re.escape(project_name) + r"\s*(?:==|>=|<=|>|<)\s*(?P<version>[^\n;]+))",
            # TODO (@NiklasRosenstein): Also match if the requirements is on the same line
            # TODO (@NiklasRosenstein): Also match extra requires
        ]
        for expr in expressions:
            refs += match_version_ref_pattern_on_lines(setup_cfg, expr)
    return refs



import typing as t

from cleo.io.null_io import NullIO  # type: ignore[import]

from slap.application import Application
from slap.check import Check, CheckResult, check, get_checks
from slap.ext.application.release import ReleaseCommandPlugin
from slap.ext.release.source_code_version import SourceCodeVersionReferencesPlugin
from slap.plugins import CheckPlugin
from slap.project import Project


class ReleaseChecksPlugin(CheckPlugin):
    """Performs some checks relevant for the `slap release` command."""

    def get_project_checks(self, project: Project) -> t.Iterable[Check]:
        return get_checks(self, project)

    def get_application_checks(self, app: Application) -> t.Iterable[Check]:
        return get_checks(self, app)

    @check("source-code-version")
    def check_packages_have_source_code_version(self, project: Project) -> tuple[CheckResult, str]:
        """Checks if all Python packages in the project have a version defined in the source code."""

        if not project.packages():
            return Check.WARNING, "No packages detected"

        matcher = SourceCodeVersionReferencesPlugin()
        matcher.io = NullIO()
        version_refs = matcher.get_version_refs(project)
        packages_without_version = {p.name for p in project.packages() or []}

        for ref in version_refs:
            for package in project.packages() or []:
                if ref.file.is_relative_to(package.path):
                    packages_without_version.discard(package.name)

        return (
            Check.ERROR if packages_without_version else Check.OK,
            (f'The following packages have no <b>__version__</b>: <b>{", ".join(packages_without_version)}</b>')
            if packages_without_version
            else f'Found <b>__version__</b> in <b>{", ".join(x.name for x in project.packages() or [])}</b>',
        )

    @check("consistent-versions")
    def check_version_number_consistency(self, app: Application) -> tuple[CheckResult, str]:
        """Checks if the version numbers in the project source code, project configuration and any other instances
        that are detected by release plugins or in the `[tool.slap.release].references` option are consistent."""

        releaser = ReleaseCommandPlugin()
        releaser.load_configuration(app)

        version_refs = releaser._get_version_refs()
        cardinality = len(set(r.value for r in version_refs))

        if cardinality == 0:
            result = Check.WARNING
            message = "No version references found"
        elif cardinality == 1:
            result = Check.OK
            message = "All version references are equal"
        else:
            result = Check.ERROR
            message = f"Found <b>{cardinality}</b> differing version references"

        return result, message



import typing as t
from pathlib import Path

import requests
from nr.util import Optional
from nr.util.fs import get_file_in_directory

from slap.check import Check, CheckResult, check, get_checks
from slap.ext.project_handlers.poetry import PoetryProjectHandler
from slap.plugins import CheckPlugin
from slap.project import Project
from slap.util.external.pypi_classifiers import get_classifiers


def get_readme_path(project: Project) -> Path | None:
    """Tries to detect the project readme. If `tool.poetry.readme` is set, that file will be returned."""

    # TODO (@NiklasRosenstein): Support other config styles that specify a readme.

    poetry: dict = project.pyproject_toml.value_or({})
    poetry = poetry.get("tool", {}).get("poetry", {})

    if (readme := poetry.get("readme")) and Path(readme).is_file():
        return Path(readme)

    return get_file_in_directory(Path.cwd(), "README", ["README.md", "README.rst", "README.txt"], case_sensitive=False)


class PoetryChecksPlugin(CheckPlugin):
    """Check plugin to validate the Poetry configuration and compare it with Slap's expectations.

    Plugin ID: `poetry`"""

    def get_project_checks(self, project: Project) -> t.Iterable[Check]:
        self.project = project
        pyproject: dict[str, t.Any] = project.pyproject_toml.value_or({})
        if isinstance(project.handler(), PoetryProjectHandler):
            self.poetry = pyproject.get("tool", {}).get("poetry")
            if self.poetry is None:
                yield Check(
                    "config", CheckResult.ERROR, "No [tool.poetry] configuration in <code>pyproject.toml</code>"
                )
                return
            yield from get_checks(self, project)

    @check("readme")
    def get_readme_check(self, project: Project) -> tuple[CheckResult, str]:
        """Checks if Poetry will be able to pick up the right readme file."""

        default_readmes = ["README.md", "README.rst"]
        detected_readme = (
            Optional(get_readme_path(self.project))
            .map(lambda p: str(p.resolve().relative_to(Path.cwd())))
            .or_else(None)
        )
        poetry_readme = self.poetry.get("readme")
        if poetry_readme is None and detected_readme in default_readmes:
            return Check.Result.OK, f"Poetry will autodetect your readme (<b>{detected_readme}</b>)"
        if poetry_readme == detected_readme:
            return Check.Result.OK, f"Poetry readme is configured correctly (path: <b>{detected_readme}</b>)"
        return (
            Check.Result.WARNING,
            f"Poetry readme appears to be misconfigured (detected: <b>{detected_readme}</b>, "
            f"configured: <b>{poetry_readme}</b>)",
        )

    @check("urls")
    def get_urls_check(self, project: Project) -> tuple[CheckResult, str]:
        """Checks if URLs are configured in the Poetry configuration and recommends to configure the `Homepage`,
        `Repository`, `Documentation` and `Bug Tracker` URLs under `[tool.poetry.urls]`."""

        has_homepage = "homepage" in self.poetry or "homepage" in {
            x.lower() for x in self.poetry.get("urls", {}).keys()
        }
        has_repository = "repository" in {x.lower() for x in self.poetry.get("urls", {}).keys()}
        has_documentation = "documentation" in {x.lower() for x in self.poetry.get("urls", {}).keys()}
        has_bug_tracker = "bug tracker" in {x.lower() for x in self.poetry.get("urls", {}).keys()}

        if has_homepage and has_repository and has_documentation and has_bug_tracker:
            return Check.OK, "Your project URLs are in top condition."
        else:
            missing = [
                k
                for k, v in {
                    "Homepage": has_homepage,
                    "Repository": has_repository,
                    "Documentation": has_documentation,
                    "Bug Tracker": has_bug_tracker,
                }.items()
                if not v
            ]
            result = Check.RECOMMENDATION if has_homepage else Check.WARNING
            message = "Please configure the following URLs: " + ", ".join(f'<s>"{k}"</s>' for k in missing)
            return result, message

    @check("classifiers")
    def get_classifiers_check(self, project: Project) -> tuple[CheckResult, str]:
        """Checks if all Python package classifiers are valid and recommends to configure them if none are set."""

        # TODO: Check for recommended classifier topics (Development State, Environment,
        #       Programming Language, Topic, Typing, etc.)
        classifiers = self.poetry.get("classifiers")  # TODO: Support classifiers in [project]
        if not classifiers:
            return Check.RECOMMENDATION, "Please configure classifiers."
        else:
            try:
                good_classifiers = get_classifiers()
            except requests.RequestException as exc:
                return Check.WARNING, f"Could not validate classifiers because list could not be fetched ({exc})"
            else:
                bad_classifiers = set(classifiers) - set(good_classifiers)
                if bad_classifiers:
                    return Check.ERROR, "Found bad classifiers: " + ",".join(f'<s>"{c}"</s>' for c in bad_classifiers)
                else:
                    return Check.OK, "All classifiers are valid."

    @check("license")
    def get_license_check(self, project: Project) -> tuple[CheckResult, str]:
        """Checks if package license is a valid SPDX license identifier and recommends to configure a license if
        none is set."""

        from slap.util.external.licenses import get_spdx_licenses

        license = self.poetry.get("license")
        if not license:
            return Check.ERROR, "Missing license"
        else:
            if license not in get_spdx_licenses():
                return Check.WARNING, f'License <s>"{license}"</s> is not a known SPDX license identifier.'
            else:
                return Check.OK, f'License <s>"{license}"</s> is a valid SPDX identifier.'



from __future__ import annotations

import logging
import typing as t
from pathlib import Path

from slap.util.toml_file import TomlFile

if t.TYPE_CHECKING:
    from nr.util.functional import Once

logger = logging.getLogger(__name__)


class Configuration:
    """Represents the configuration stored in a directory, which is either read from `slap.toml` or `pyproject.toml`."""

    #: The directory of the project. This is the directory where the `slap.toml` or `pyproject.toml` configuration
    #: would usually reside in, but the existence of neither is absolutely required.
    directory: Path

    #: Points to the `pyproject.toml` file in the project and can be used to conveniently check for its existence
    #: or to access its contents.
    pyproject_toml: TomlFile

    #: Points to the `slap.toml` file in the project and can be used to conveniently check for its existence
    #: or to access its contents.
    slap_toml: TomlFile

    #: Use this to access the Slap configuration, automatically loaded from either `slap.toml` or the `tool.slap`
    #: section in `pyproject.toml`. The attribute is a #Once instance, thus it needs to be called to retrieve
    #: the contents. This is the same as #get_raw_configuration(), but is more efficient.
    raw_config: Once[dict[str, t.Any]]

    def __init__(self, directory: Path) -> None:
        from nr.util.functional import Once

        self.directory = directory
        self.pyproject_toml = TomlFile(directory / "pyproject.toml")
        self.slap_toml = TomlFile(directory / "slap.toml")
        self.raw_config = Once(self.get_raw_configuration)

    def __repr__(self) -> str:
        return f'{type(self).__name__}(directory="{self.directory}")'

    def get_raw_configuration(self) -> dict[str, t.Any]:
        """Loads the raw configuration data for Slap from either the `slap.toml` configuration file or `pyproject.toml`
        under the `[slap.tool]` section. If neither of the files exist or the section in the pyproject does not exist,
        an empty dictionary will be returned."""

        if self.slap_toml.exists():
            logger.debug("Reading configuration for <subj>%s</subj> from <val>%s</val>", self, self.slap_toml.path)
            return self.slap_toml.value()
        if self.pyproject_toml.exists():
            logger.debug("Reading configuration for <subj>%s</subj> from <val>%s</val>", self, self.pyproject_toml.path)
            return self.pyproject_toml.value().get("tool", {}).get("slap", {})
        return {}



import typing as t


def toml_highlight(toml_data: dict[str, t.Any] | str) -> str:
    import pygments
    import pygments.formatters
    import pygments.lexers
    import tomli_w

    if not isinstance(toml_data, str):
        toml_data = tomli_w.dumps(toml_data)
    return pygments.highlight(
        toml_data, pygments.lexers.get_lexer_by_name("toml"), pygments.formatters.get_formatter_by_name("terminal")
    )



"""
An API to describe dependency specifications in a Python project.

References:

* [PEP 440][]
* [PEP 508][]
* [Poetry Dependencies][]

[PEP 440]: https://peps.python.org/pep-0440/
[PEP 508]: https://peps.python.org/pep-0508/
[Poetry Dependencies]: https://python-poetry.org/docs/dependency-specification/
"""

from __future__ import annotations

import dataclasses
import re
import typing as t
from pathlib import Path
from urllib.parse import parse_qs, parse_qsl, urlparse, urlunparse

from nr.util.generic import T


class VersionSpec:
    """Represents a version specification, which is either a [PEP 440][] version number, a [PEP 508][]
    dependency specification, or a [Poetry Dependencies][] specification string."""

    def __init__(self, version_spec: str) -> None:
        from poetry.core.packages.dependency import Dependency as _PoetryDependency  # type: ignore[import]

        self.__original = version_spec.strip()
        self.__dependency = _PoetryDependency("", self.__original)

    def __bool__(self) -> bool:
        """Returns `True` if the version spec is initialized from an empty string. Note that it will otherwise
        behave exactly the same as if it was initialized with `*` (matching any package version)."""

        return bool(self.__original) and self.__original != "*"

    def __str__(self) -> str:
        return self.__original

    def __repr__(self) -> str:
        return f"VersionSpec({self.__original!r})"

    def __eq__(self, other: t.Any) -> bool:
        if isinstance(other, VersionSpec):
            return self.__dependency == other.__dependency and bool(self) == bool(other)
        return False

    def to_pep_508(self) -> str:
        # NOTE (@NiklasRosenstein): Removes parentheses around the spec.
        return self.__dependency.to_pep_508().strip()[1:-1]

    def accepts(self, version: str) -> bool:
        """Tests if the version spec accepts the given version string."""

        from poetry.core.semver.version import Version  # type: ignore[import]

        return self.__dependency.constraint.allows(Version.parse(version))


@dataclasses.dataclass
class Dependency:
    """Base data model for dependency specifications."""

    #: The dependency name.
    name: str

    #: A list of extras to install.
    extras: list[str] | None = None

    #: A [PEP 440][] version or dependency specification or a [Poetry Dependencies][] specification that
    #: specifies the range of Python versions that this dependency should be installed for.
    python: VersionSpec | None = None

    #: A [PEP 508][] environment marker that gives more flexibility as to the conditions that the dependency
    #: should be installed.
    markers: str | None = None

    #: A list of hashes that the installed dependencies' package must match. Each value must be a combined string
    #: of the hash algorithm with the hash value separated by a colon. Note that this may not be supported by all
    #: dependency types (e.g. #GitDependency).
    hashes: list[str] | None = None


@dataclasses.dataclass
class _PypiDependency:

    name: str

    #: The version specification for the package.
    version: VersionSpec

    #: The alias of a PyPI index to install the package from.
    source: str | None = None


@dataclasses.dataclass
class _GitDependency:

    name: str

    #: The repository URL to get the Python package from.
    url: str

    #: The revision of the repository. Should not be combined with #branch or #tag.
    rev: str | None = None

    #: The branch of the repository. Should not be combined with #branch or #tag.
    branch: str | None = None

    #: The tag of the repository. Should not be combined with #branch or #tag.
    tag: str | None = None


@dataclasses.dataclass
class _PathDependency:

    name: str

    #: The path from which to install the Python package from.
    path: Path

    #: Whether the package should be installed in development mode.
    develop: bool = False

    #: Whether the package should be symlinked.
    link: bool = False


@dataclasses.dataclass
class _UrlDependency:

    name: str

    #: The URL to get the package to install from.
    url: str


@dataclasses.dataclass
class _MultiDependency:

    name: str

    #: Defines multiple possible ways to install a dependency. Each dependency should differ in their #Dependency.python
    #: and/or #Dependency.marker specification.
    dependencies: list[Dependency]


@dataclasses.dataclass
class PypiDependency(Dependency, _PypiDependency):
    """A dependency on a package in a Python Package Index."""

    @staticmethod
    def parse(value: str) -> PypiDependency:
        """Parses a package name and its version spec from a string."""

        value, markers = value.partition(";")[::2]

        match = re.match(r"\s*[^<>=!~\^\(\)\*]+", value)
        if match:
            name = match.group(0)
            version_spec = VersionSpec(value[match.end() :].strip() or "*")  # noqa: E203
        else:
            name = value
            version_spec = VersionSpec("")

        name, extras = split_package_name_with_extras(name)
        return PypiDependency(name=name, version=version_spec, extras=extras, markers=markers.strip() or None)

    @staticmethod
    def parse_list(lst: t.Iterable[str]) -> list[PypiDependency]:
        """Parses a list of strings as #PypiDependency#s."""

        return [PypiDependency.parse(x) for x in lst]


@dataclasses.dataclass
class GitDependency(Dependency, _GitDependency):
    """A dependency on a Python package that can be installed from a Git repository."""


@dataclasses.dataclass
class PathDependency(Dependency, _PathDependency):
    """A dependency on a Python package that can be installed from a Path."""


@dataclasses.dataclass
class UrlDependency(Dependency, _UrlDependency):
    """A dependency on a Python package that can be installed from a URL."""


@dataclasses.dataclass
class MultiDependency(Dependency, _MultiDependency):
    """Express multiple possible ways to install a Python package."""


#: Represents a dependency configuration that does not contain the dependency name, such as is used for example
#: by Poetry. A plain string is parsed like a dependency string (see #parse_dependency_string()), while a dictionary
#: represents a more complex dependency configuration that can be parsed into other types of dependencies as well.
DependencyConfig = str | dict[str, t.Any] | list[dict[str, t.Any]]


def split_package_name_with_extras(value: str) -> tuple[str, list[str] | None]:
    """Splits *value* as a string that contains a package name and optionally its extras into components."""

    match = re.match(r"\s*([^\[\]]+?)?\s*(?:\[([^\[\]]+)\])?\s*$", value)
    if not match:
        raise ValueError(f"invalid package name with extras: {value!r}")

    extras = [x.strip() for x in match.group(2).split(",")] if match.group(2) else None
    if extras is not None and not all(extras):
        raise ValueError(f"invalid package name with extras: {value!r}")

    return match.group(1), extras


def parse_dependency_string(value: str) -> Dependency:
    """
    Convert *value* to a representation as a #Dependency subclass.

    * In addition to the [PEP 508][] dependency specification, the function supports a `--hash` option as is also
      supported by Pip. Hashes in URL fragments are also parsed into #Dependency.hashes.
    * URL formatted specifications can also be Git repository URLs or paths (must be in Posix format as absolute
      path, or an explicit relative path, i.e. begin with curdir or pardir).

    Args:
      value: The dependency specification.
    Raises:
      ValueError: If the string cannot be parsed into a #Dependency.

    !!! note A URL or Git dependency must still contain a package name (i.e. be of the form `<name> @ <url>`). If
      a URL or Git repository URL is encountered without a package name, a #ValueError is raised.
    """

    value = value.strip()
    if value.startswith("http://") or value.startswith("https://") or value.startswith("git+"):
        raise ValueError(f"A plain URL or Git repository URL must be prefixed with a package name: {value!r}")

    # Extract trailing options from the dependency.
    hashes: list[str] = []

    def handle_option(match: re.Match) -> str:
        if match.group(1) == "hash":
            hashes.append(match.group(2))
        return ""

    value = re.sub(r"\s--(\w+)=(.*)(\s|$)", handle_option, value)

    # Check if it's a dependency of the form `<name> @ <package>`. This can be either a
    # #UrlDependency or #GitDependency.
    if "@" in value:
        markers: str | None
        name, url = value.partition("@")[::2]
        name, extras = split_package_name_with_extras(name)
        url, markers = url.partition(";")[::2]
        markers = markers.strip() or None
        urlparts = urlparse(url.strip())

        # Remove the fragments from the URL.
        url = urlunparse((urlparts.scheme, urlparts.netloc, urlparts.path, urlparts.params, urlparts.query, None))

        # Parse it as a Git URL.
        if url.startswith("git+"):

            def unpack(val: t.Sequence[T] | None) -> T | None:
                return val[0] if val else None

            options = parse_qs(urlparts.fragment)
            return GitDependency(
                name=name,
                url=url[4:],
                rev=unpack(options.get("rev")),
                branch=unpack(options.get("branch")),
                tag=unpack(options.get("tag")),
                extras=extras,
                markers=markers,
                hashes=hashes or None,
            )

        # Parse it as a path.
        elif url.startswith("/") or url.startswith("./") or url.startswith("../"):
            options = parse_qs(urlparts.fragment)
            return PathDependency(
                name=name,
                path=Path(url),
                develop="develop" in options,
                link="link" in options,
                extras=extras,
                markers=markers,
                hashes=hashes or None,
            )

        elif urlparts.scheme:
            # Treat all fragments as hash options.
            hashes += [f"{item[0]}:{item[1]}" for item in parse_qsl(urlparts.fragment)]
            return UrlDependency(
                name=name,
                url=url,
                extras=extras,
                markers=markers,
                hashes=hashes or None,
            )

        else:
            raise ValueError(f"invalid URL-formatted dependency: {value!r}")

    # TODO (@NiklasRosenstein): Support parsing path dependencies.

    dependency = PypiDependency.parse(value)
    dependency.hashes = hashes or None
    return dependency


def _parse_single_dependency_config(name: str, dep: str | dict[str, t.Any]) -> Dependency:
    """Convert a single dependency expressed as a table or dictionary to a Slap dependency specification.

    Examples:

        >>> _parse_single_dependency_config('foo', 'git+https://github.com/foo/foo.git') == GitDependency('foo', 'https://github.com/foo/foo.git')
        >>> _parse_single_dependency_config('foo', {'version': '1.2.3'}) == PypiDependency('foo', VersionSpec('1.2.3'))
        >>> _parse_single_dependency_config('foo', {'git': 'https://...'}) == GitDependency('foo', 'https://...')
    """  # noqa: E501

    dependency: Dependency

    if isinstance(dep, str):
        dep = dep.strip()

        # Check if the dependency specification appears to be in URL format.
        if (
            dep.startswith("git+")
            or dep.startswith("http://")
            or dep.startswith("https://")
            or dep.startswith("../")
            or dep.startswith("./")
            or dep.startswith("/")
        ):
            dependency = parse_dependency_string(f"{name} @ {dep}")
        else:
            # If dep is _just_ a version number, we need to prefix it with a = to ensure the PypiDependency
            # is parsed correctly.
            dep = dep.strip()
            if dep and dep[0].isnumeric():
                dep = f"={dep}"
            dependency = parse_dependency_string(f"{name} {dep}")

    elif "git" in dep:
        dependency = GitDependency(
            name=name, url=dep["git"], rev=dep.get("rev"), branch=dep.get("branch"), tag=dep.get("tag")
        )

    elif "path" in dep:
        # NOTE (@NiklasRosenstein): The "link" key is actually not a Poetry feature, but it won't complain if
        #   you are just using Slap anyway.
        dependency = PathDependency(
            name=name, path=Path(dep["path"]), develop=dep.get("develop", False), link=dep.get("link", False)
        )

    elif "url" in dep:
        dependency = UrlDependency(name=name, url=dep["url"])

    elif "version" in dep:
        dependency = PypiDependency(
            name=name,
            version=VersionSpec(dep["version"]),
            source=dep.get("source"),
        )

    else:
        raise ValueError(f"Cannot interpret dependency: {name} = {dep!r}")

    if not isinstance(dep, str):
        dependency.python = VersionSpec(dep["python"]) if dep.get("python") else None
        dependency.markers = dep.get("markers")

    return dependency


def parse_dependency_config(name: str, dep: DependencyConfig) -> Dependency:
    """Converts a dependency configuration. This "happens" to be compatible with the Poetry configuration format."""

    if isinstance(dep, dict | str):
        dependency = _parse_single_dependency_config(name, dep)
    elif isinstance(dep, list):
        dependency = MultiDependency(name, [_parse_single_dependency_config(name, x) for x in dep])
    else:
        raise ValueError(f"Cannot interpret dependency: {name} = {dep!r}")

    return dependency


def parse_dependencies(dependencies: list[str] | dict[str, DependencyConfig]) -> list[Dependency]:
    """Converts Poetry dependencies to Slap dependencies.

    Args:
      dependencies: Either a list of strings (for example used in the `tool.poetry.extras` section where you just
        specify a list of strings that contain the package name and version spec as one), or a dictionary mapping
        the dependency name to its specification. The specification can be a list of specifications to describe a
        #MultiDependency.
    """

    if isinstance(dependencies, list):
        return [parse_dependency_string(dep) for dep in dependencies]

    elif isinstance(dependencies, dict):
        return [parse_dependency_config(key, value) for key, value in dependencies.items()]

    else:
        raise TypeError(type(dependencies))



import os
import re
import subprocess as sp
from pathlib import Path

from slap.plugins import ChangelogUpdateAutomationPlugin


class GithubActionsChangelogUpdateAutomationPlugin(ChangelogUpdateAutomationPlugin):
    """A plugin for use in GitHub Actions via `slap changelog update-pr --use github-actions` which will do all steps
    to push the updated changelogs back to a pull request branch. It should be used only in an action that is run as
    part of a GitHub pull request.

    The GitHub Action should be configured like this:

    ```yaml
    update-changelogs:
      if: github.event_name == 'pull_request'
      permissions:
        contents: write
      runs-on: ubuntu-latest
      steps:
      - uses: actions/checkout@v2
      - name: Set up Python 3.10
        uses: actions/setup-python@v2
        with: {{ python-version: '3.10' }}
      - name: Install Slap
        run: pip install slap-cli==1.0.2
      - name: Update PR references in changelogs
        run: slap changelog update-pr --use github-actions
    ```

    GitHub environment variables used:

    * `GITHUB_HEAD_REF`
    * `GITHUB_BASE_REF`
    * `GITHUB_REF` (the github PR number formatted as `refs/pull/{id}`)

    Additional environment variables to control the plugin:

    * `GIT_USER_NAME` (defaults to `GitHub Action`)
    * `GIT_USER_EMAIL` (defaults to `github-action@users.noreply.github.com`)
    * `GIT_COMMIT_MESSAGE` (defaults to `Update changelog PR references`)
    * `GIT_SHOW_DIFF` (if set, runs "git diff" before publishing changes)
    """

    def initialize(self) -> None:
        sp.check_output(["git", "fetch", "origin", os.environ["GITHUB_BASE_REF"]], stderr=sp.PIPE)
        # sp.check_output(['git', 'checkout', 'origin/' + os.environ['GITHUB_HEAD_REF']], stderr=sp.PIPE)
        # sp.check_output(['git', 'checkout', '-b', os.environ['GITHUB_HEAD_REF']], stderr=sp.PIPE)

    def get_base_ref(self) -> str:
        return "origin/" + os.environ["GITHUB_BASE_REF"]

    def get_pr(self) -> str:
        ref = os.environ["GITHUB_REF"]
        match = re.match(r"refs/pull/(\d+)", ref)
        if not match:
            raise EnvironmentError(f'could not determine Pull Request ID from GITHUB_REF="{ref}"')
        return match.group(1)

    def publish_changes(self, changed_files: list[Path]) -> None:
        user_name = os.environ.get("GIT_USER_NAME", "GitHub Action")
        user_email = os.environ.get("GIT_USER_EMAIL", "github-action@users.noreply.github.com")
        commit_message = os.environ.get("GIT_COMMIT_MESSAGE", "Update changelog PR references")
        if os.getenv("GIT_SHOW_DIFF"):
            sp.check_output(["git", "diff"], stderr=sp.PIPE)
        sp.check_output(["git", "add"] + [str(f) for f in changed_files], stderr=sp.PIPE)
        sp.check_output(
            ["git", "-c", "user.name=" + user_name, "-c", "user.email=" + user_email, "commit", "-m", commit_message],
            stderr=sp.PIPE,
        )
        sp.check_output(["git", "push", "origin", os.environ["GITHUB_HEAD_REF"]], stderr=sp.PIPE)



""" Project handler for projects using the Poetry build system. """

from __future__ import annotations

import logging
import typing as t

from slap.ext.project_handlers.base import PyprojectHandler
from slap.project import Dependencies, Package, Project

if t.TYPE_CHECKING:
    from slap.python.dependency import Dependency

logger = logging.getLogger(__name__)


class PoetryProjectHandler(PyprojectHandler):

    # ProjectHandlerPlugin

    def matches_project(self, project: Project) -> bool:
        if not project.pyproject_toml.exists():
            return False
        build_backend = project.pyproject_toml.get("build-system", {}).get("build-backend")
        return build_backend == "poetry.core.masonry.api"

    def get_dist_name(self, project: Project) -> str | None:
        return project.pyproject_toml.get("tool", {}).get("poetry", {}).get("name")

    def get_readme(self, project: Project) -> str | None:
        return project.pyproject_toml.get("tool", {}).get("poetry", {}).get("readme") or super().get_readme(project)

    def get_packages(self, project: Project) -> list[Package] | None:
        packages = project.pyproject_toml.get("tool", {}).get("poetry", {}).get("packages")
        if packages is None:
            return super().get_packages(project)  # Fall back to automatically determining the packages
        if not packages:
            return None  # Indicate explicitly that the project does not expose packages

        return [
            Package(
                name=p["include"].replace("/", "."),
                path=project.directory / p.get("from", "") / p["include"],
                root=project.directory / p.get("from", ""),
            )
            for p in packages
        ]

    def get_dependencies(self, project: Project) -> Dependencies:
        from slap.install.installer import Indexes
        from slap.python.dependency import PypiDependency, parse_dependencies

        poetry: dict[str, t.Any] = project.pyproject_toml.get("tool", {}).get("poetry", {})
        dependencies = parse_dependencies(poetry.get("dependencies", []))
        python = next((d for d in dependencies if d.name == "python"), None)
        if python is not None:
            assert isinstance(python, PypiDependency), repr(python)

        # Collect the package indexes from the Poetry config.
        indexes = Indexes()
        for source in poetry.get("source", []):
            if source.get("default", False):
                indexes.default = source["name"]
            indexes.urls[source["name"]] = source["url"]

        return Dependencies(
            python=python.version if python else None,
            run=[d for d in dependencies if d.name != "python"],
            dev=parse_dependencies(poetry.get("dev-dependencies", [])),
            extra={k: parse_dependencies(v) for k, v in poetry.get("extras", {}).items()},
            build=PypiDependency.parse_list(project.pyproject_toml.get("build-system", {}).get("requires", [])),
            indexes=indexes,
        )

    def get_add_dependency_toml_location_and_config(
        self,
        project: Project,
        dependency: Dependency,
        where: str,
    ) -> tuple[list[str], list | dict]:
        from slap.python.dependency import PypiDependency

        if not isinstance(dependency, PypiDependency):
            raise Exception(f"Poetry project handler only supports PypiDependency, got {dependency!r}")

        locator = ["dependencies"] if where == "run" else ["dev-dependencies"] if where == "dev" else ["extras", where]
        value: list | dict = (
            {dependency.name: convert_dependency_to_poetry_config(dependency)}
            if where in ("run", "dev")
            else [f"{dependency.name} {dependency.version}"]
        )
        return ["tool", "poetry"] + locator, value


def convert_dependency_to_poetry_config(dependency: Dependency) -> t.Mapping[str, t.Any] | str:
    import tomlkit.api

    from slap.python.dependency import PypiDependency

    if isinstance(dependency, PypiDependency):
        if not dependency.markers and not dependency.python and not dependency.source:
            return str(dependency.version)
        result = tomlkit.api.inline_table()
        result["version"] = str(dependency.version)
        if dependency.markers:
            result["markers"] = dependency.markers
        if dependency.python:
            result["python"] = str(dependency.python)
        if dependency.extras:
            result["extras"] = dependency.extras
        if dependency.source:
            result["source"] = dependency.source
        return result
    else:
        raise ValueError(f"currently only supports PypiDependency, got {dependency!r}")



""" Provides an installer for dependencies. """

from __future__ import annotations

import abc
import dataclasses
import logging
import shlex
import subprocess as sp
import typing as t
from pathlib import Path
from urllib.parse import unquote

from nr.util.url import Url

from slap.python.dependency import MultiDependency
from slap.python.pep508 import filter_dependencies, test_dependency

if t.TYPE_CHECKING:
    from slap.project import Project
    from slap.python.dependency import Dependency
    from slap.python.environment import PythonEnvironment

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class IndexSpec:
    name: str
    url: str | None
    username: str | None
    password: str | None

    @property
    def url_with_auth(self) -> str:
        assert self.url is not None, "IndexSpec.url is None"
        parsed_url = Url.of(self.url)
        if self.username:
            parsed_url.username = self.username
        if self.password:
            parsed_url.password = self.password
        return str(parsed_url)

    @classmethod
    def parse(cls, spec: str) -> IndexSpec:
        """Parses a spec for an extra index URL which must be of the form `name=...,url=https://...` and may
        additional provide values a `username=...` and `password=...`."""

        values = {k.strip(): unquote(v.strip()) for k, v in (x.partition("=")[::2] for x in spec.split(","))}
        try:
            name = values.pop("name")
            url = values.pop("url", None)
            username = values.pop("username", None)
            password = values.pop("password", None)
        except KeyError as exc:
            raise ValueError(f"invalid index spec {spec!r}: missing {exc}")
        for key in values:
            raise ValueError(f"invalid index spec {spec!r}: unrecognized key {key!r}")
        if not any((url, username, password)):
            raise ValueError(f"invalid index spec {spec!r}: need one of url, username, password")
        return cls(name, url, username, password)


@dataclasses.dataclass
class Indexes:
    """Represents a configuration of PyPI indexes."""

    #: The name of the default index in the #urls mapping.
    default: str | None = None

    #: A mapping that assigns each key (the name of the index) its index URL.
    urls: dict[str, str] = dataclasses.field(default_factory=dict)

    def combine_with(self, other: Indexes) -> None:
        """Combine this configuration with another. All values from *self* take precedence."""

        if other.default and self.default and other.default != self.default:
            logger.warning(
                "Conflicting default index between projects in repository: %r (current), %r",
                self.default,
                other.default,
            )
        self.default = self.default or other.default

        # TODO (@NiklasRosenstein): Warn about conflicting package indexes.
        self.urls = {**self.urls, **other.urls}


@dataclasses.dataclass
class InstallOptions:
    indexes: Indexes
    quiet: bool
    upgrade: bool


class Installer(abc.ABC):
    """An installer for dependencies into a #PythonEnvironment."""

    @abc.abstractmethod
    def install(self, dependencies: list[Dependency], target: PythonEnvironment, options: InstallOptions) -> int:
        ...


class SymlinkHelper(t.Protocol):
    """
    A helper for introspecting a project for additional dependencies and symlinking it. This is needed when a
    #PathDependency is encountered with #PathDependency.link enabled.
    """

    def get_dependencies_for_project(self, project: Path) -> list[Dependency]:
        ...

    def link_project(self, project: Path) -> None:
        ...


class PipInstaller(Installer):
    """Installs dependencies via Pip."""

    def __init__(self, symlink_helper: SymlinkHelper | None) -> None:
        """
        Args:
          symlink_helper: A helper for implementing #PathDependency.link when it is encountered. If not specified,
            an error will be raised when a #PathDependency is passed that needs to be linked.
        """

        self.symlink_helper = symlink_helper

    def install(self, dependencies: t.Sequence[Dependency], target: PythonEnvironment, options: InstallOptions) -> int:

        from slap.python.dependency import PathDependency, PypiDependency, UrlDependency

        # Collect the Pip arguments and the dependencies that need to be installed through other methods.
        supports_hashes = {PypiDependency, UrlDependency}
        unsupported_hashes: dict[type[Dependency], list[Dependency]] = {}
        link_projects: list[Path] = []
        pip_arguments: list[str] = []
        # used_indexes: set[str] = set()
        dependencies = list(dependencies)

        while dependencies:
            dependency = dependencies.pop()

            # TODO (@NiklasRosenstein): Pass extras from PipInstaller caller.
            if not test_dependency(dependency, target.pep508, set()):
                continue

            # Collect dependencies for which hashes are not supported so we can report it later.
            if dependency.hashes and type(dependency) not in supports_hashes:
                unsupported_hashes.setdefault(type(dependency), []).append(dependency)

            if isinstance(dependency, PathDependency) and dependency.link:
                logger.info("Collecting recursive dependencies for project <val>%s</val>", dependency.path)
                if self.symlink_helper is None:
                    raise Exception(
                        "Unable to install %r because no symlink helper is available in this context", dependency
                    )
                dependencies += filter_dependencies(
                    dependencies=self.symlink_helper.get_dependencies_for_project(dependency.path),
                    env=target.pep508,
                    extras=set(dependency.extras or []),
                )
                link_projects.append(dependency.path)
                continue

            if isinstance(dependency, MultiDependency):
                for sub_dependency in dependency.dependencies:
                    # TODO (@NiklasRosenstein): Pass extras from the caller so we can evaluate them here
                    if test_dependency(sub_dependency, target.pep508, set()):
                        dependencies.insert(0, sub_dependency)

            else:
                pip_arguments += self.dependency_to_pip_arguments(dependency)

            # if isinstance(dependency, PypiDependency) and dependency.source:
            #     used_indexes.add(dependency.source)

        # Add the extra index URLs.
        # TODO (@NiklasRosenstein): Inject credentials for index URLs.
        # NOTE (@NiklasRosenstein): While the dependency configuration allows you to specify exactly for each
        #   dependency where it should be fetched from, with the Pip CLI we cannot currently have that level
        #   of control.
        try:
            if options.indexes.default is not None:
                pip_arguments += ["--index-url", options.indexes.urls[options.indexes.default]]
            # for index_name in used_indexes - {options.indexes.default}:
            # NOTE (@NiklasRosenstein): For now we just pass all indexes to Pip. When you run `slap install` without
            #       the `--link` option, the package will be installed directly with Pip, thus the runtime dependencies
            #       are not passed here and we would not recognize the extra indexes required for those dependencies.
            for index_name in options.indexes.urls.keys() - {options.indexes.default}:
                pip_arguments += ["--extra-index-url", options.indexes.urls[index_name]]
        except KeyError as exc:
            raise Exception(f"PyPI index {exc} is not configured")

        # Construct the Pip command to run.
        pip_command = [target.executable, "-m", "pip", "install"] + pip_arguments
        if options.quiet:
            pip_command += ["-q"]
        if options.upgrade:
            pip_command += ["--upgrade"]

        logger.info("Installing with Pip using command <subj>$ %s</subj>", " ".join(map(shlex.quote, pip_command)))
        if (res := sp.call(pip_command)) != 0:
            return res

        # Symlink all projects that need to be linked.
        for project_path in link_projects:
            assert self.symlink_helper is not None
            self.symlink_helper.link_project(project_path)

        return 0

    @staticmethod
    def dependency_to_pip_arguments(dependency: Dependency) -> list[str]:
        """Converts a dependency to a list of arguments for Pip.

        Args:
          dependency: The dependency to convert. Must be one of #GitDependency, #PathDependency,
            #PypiDependency or #UrlDependency. A #MultiDependency is not supported by this function.
        Raises:
          Exception: If an unexpected kind of dependency was encountered.
        """

        from slap.python.dependency import GitDependency, PathDependency, PypiDependency, UrlDependency

        extras = "" if not dependency.extras else f'[{",".join(dependency.extras)}]'
        hashes = " ".join(f"--hash={h}" for h in dependency.hashes or [])
        pip_arguments = []

        if isinstance(dependency, GitDependency):
            # TODO (@NiklasRosenstein): Add Git branch/rev/tag to the URL.
            if dependency.branch or dependency.rev or dependency.tag:
                logger.warning(
                    "PipInstaller does not currently support Git branch/rev/tag, dependency will be installed "
                    "from main branch: <val>%s</val>",
                    dependency,
                )
            pip_arguments += [f"{dependency.name}{extras} @ git+{dependency.url}"]

        elif isinstance(dependency, PathDependency):
            assert not dependency.link  # We caught that case before
            if dependency.develop:
                pip_arguments += ["-e"]
            prefix = "" if dependency.path.is_absolute() else "./"
            pip_arguments += [f"{prefix}{dependency.path}{extras}"]

        elif isinstance(dependency, PypiDependency):
            pip_arguments += [f"{dependency.name}{extras} {dependency.version.to_pep_508()} {hashes}".rstrip()]

        elif isinstance(dependency, UrlDependency):
            pip_arguments += [f"{dependency.name}{extras} @ {dependency.url} {hashes}".rstrip()]

        else:
            raise Exception(f"Unexpected dependency type: {dependency}")

        assert pip_arguments, dependency
        return pip_arguments


def get_indexes_for_projects(projects: t.Sequence[Project]) -> Indexes:
    """Combines the indexes configuration from each project into one index."""

    indexes = Indexes()
    for project in projects:
        indexes.combine_with(project.dependencies().indexes)
    return indexes



""" With the application object we manage the CLI commands and other types of plugins as well as access to the Slap
user and project configuration. """

from __future__ import annotations

import dataclasses
import logging
import textwrap
import typing as t
from pathlib import Path

from cleo.application import Application as BaseCleoApplication  # type: ignore[import]
from cleo.commands.command import Command as _BaseCommand  # type: ignore[import]
from cleo.helpers import argument, option  # type: ignore[import]
from cleo.io.io import IO  # type: ignore[import]
from databind.core.settings import Alias

from slap import __version__

if t.TYPE_CHECKING:
    from nr.util.functional import Once

    from slap.configuration import Configuration
    from slap.project import Project
    from slap.repository import Repository

__all__ = ["Command", "argument", "option", "IO", "Application"]
logger = logging.getLogger(__name__)


class Command(_BaseCommand):
    def __init_subclass__(cls) -> None:
        if not cls.help:
            first_line, remainder = (cls.__doc__ or "").partition("\n")[::2]
            cls.help = (first_line.strip() + "\n" + textwrap.dedent(remainder)).strip()
        cls.description = cls.description or (cls.help.strip().splitlines()[0] if cls.help else None)

        # TODO (@NiklasRosenstein): Implement automatic wrapping of description text, but we
        #   need to ignore HTML tags that are used to colour the output.

        # argument: Argument
        # for argument in cls.arguments:
        #   print(argument)
        #   argument._description = '\n'.join(textwrap.wrap(argument._description or '', 70))

        # option: Option
        # for option in cls.options:
        #   print(option)
        #   option._description = '\n'.join(textwrap.wrap(option._description or '', 70))


class CleoApplication(BaseCleoApplication):

    from cleo.formatters.style import Style  # type: ignore[import]
    from cleo.io.inputs.input import Input  # type: ignore[import]
    from cleo.io.outputs.output import Output  # type: ignore[import]

    _styles: dict[str, Style]

    def __init__(self, init: t.Callable[[IO], t.Any], name: str = "console", version: str = "") -> None:
        super().__init__(name, version)
        self._init_callback = init
        self._styles = {}

        self._initialized = True
        from slap.util.cleo import HelpCommand

        self.add(HelpCommand())
        self._default_command = "help"

        self.add_style("code", "dark_gray")
        self.add_style("warning", "magenta")
        self.add_style("u", options=["underline"])
        self.add_style("i", options=["italic"])
        self.add_style("s", "yellow")
        self.add_style("opt", "cyan", options=["italic"])

    def add_style(self, name, fg=None, bg=None, options=None):
        self._styles[name] = self.Style(fg, bg, options)

    def create_io(
        self, input: Input | None = None, output: Output | None = None, error_output: Output | None = None
    ) -> IO:
        from slap.util.cleo import add_style

        io = super().create_io(input, output, error_output)
        for style_name, style in self._styles.items():
            add_style(io, style_name, style)
        return io

    def render_error(self, error: Exception, io: IO) -> None:
        import subprocess as sp

        if isinstance(error, sp.CalledProcessError):
            msg = "Uncaught CalledProcessError raised for command <subj>%s</subj> (exit code: <val>%s</val>)."
            args: tuple[t.Any, ...] = (error.args[1], error.returncode)
            stdout: str | None = error.stdout.decode() if error.stdout else None
            stderr: str | None = error.stderr.decode() if error.stderr else None
            if stdout:
                msg += "\n  stdout:\n<fg=black;attr=bold>%s</fg>"
                stdout = textwrap.indent(stdout, "    ")
                args += (stdout,)
            if stderr:
                msg += "\n  stderr:\n<fg=black;attr=bold>%s</fg>"
                stderr = textwrap.indent(stderr, "    ")
                args += (stderr,)

            logger.error(msg, *args)

        return super().render_error(error, io)

    def _configure_io(self, io: IO) -> None:
        import logging

        from nr.util.logging.formatters.terminal_colors import TerminalColorFormatter

        fmt = "<fg=bright black>%(message)s</fg>"
        if io.input.has_parameter_option("-vvv"):
            fmt = "<fg=bright black>%(asctime)s | %(levelname)s | %(name)s | %(message)s</fg>"
            level = logging.DEBUG
        elif io.input.has_parameter_option("-vv"):
            level = logging.DEBUG
        elif io.input.has_parameter_option("-v"):
            level = logging.INFO
        elif io.input.has_parameter_option("-q"):
            level = logging.ERROR
        elif io.input.has_parameter_option("-qq"):
            level = logging.CRITICAL
        else:
            level = logging.WARNING

        logging.basicConfig(level=level)
        formatter = TerminalColorFormatter(fmt)
        assert formatter.styles
        formatter.styles.add_style("subj", "blue")
        formatter.styles.add_style("obj", "yellow")
        formatter.styles.add_style("val", "cyan")
        formatter.install("tty")
        formatter.install("notty")  # Hack for now to enable it also in CI

        super()._configure_io(io)
        self._init_callback(io)

    def _run_command(self, command: Command, io: IO) -> int:
        return super()._run_command(command, io)


@dataclasses.dataclass
class ApplicationConfig:
    #: A list of application plugins to _not_ activate.
    disable: list[str] = dataclasses.field(default_factory=list)

    #: A list of plugins to enable only, causing the default plugins to not be loaded.
    enable_only: t.Annotated[list[str] | None, Alias("enable-only")] = None


class Application:
    """The application object is the main hub for command-line interactions. It is responsible for managing the project
    that is the main subject of the command-line invokation (or multiple of such), provide the #cleo command-line
    application that #ApplicationPlugin#s can register commands to, etc."""

    main_project: Once[Project | None]

    #: The application configuration loaded once via #get_application_configuration().
    config: Once[ApplicationConfig]

    #: The cleo application to which new commands can be registered via #ApplicationPlugin#s.
    cleo: CleoApplication

    def __init__(self, directory: Path | None = None, name: str = "slap", version: str = __version__) -> None:
        from nr.util.functional import Once

        self._directory = directory or Path.cwd()
        self._repository: t.Optional[Repository] = None
        self._plugins_loaded = False
        self.config = Once(self._get_application_configuration)
        self.cleo = CleoApplication(self._cleo_init, name, version)
        self.main_project = Once(self._get_main_project)

    @property
    def repository(self) -> Repository:
        """Return the Slap repository that is the subject of the current application. There may be command plugins
        that do not require the repository to function, so this property creates the repository lazily."""

        from slap.repository import Repository

        if self._repository is None:
            self._repository = Repository(self._directory)

        return self._repository

    def _get_application_configuration(self) -> ApplicationConfig:
        """Loads the application-level configuration."""

        from databind.core.settings import ExtraKeys
        from databind.json import load

        raw_config = self.repository.raw_config().get("application", {})
        return load(raw_config, ApplicationConfig, settings=[ExtraKeys(True)])

    def _get_main_project(self) -> Project | None:
        """Returns the main project, which is the one that the current working directory is pointing to."""

        cwd = Path.cwd()

        for project in self.repository.projects():
            path = project.directory.resolve()
            if path == cwd:
                return project

        return None

    def configurations(self) -> list[Configuration]:
        """Return a list of all configuration objects, i.e. all projects and eventually the #Repository, unless one
        project is from the same directory as the repository."""

        result: list[Configuration] = list(self.repository.projects())
        if self.repository.directory not in tuple(p.directory for p in self.repository.projects()):
            result.insert(0, self.repository)
        return result

    def load_plugins(self) -> None:
        """Loads all application plugins (see #ApplicationPlugin) and activates them.

        By default, all plugins available in the `slap.application.ApplicationPlugin` entry point group are loaded. This
        behaviour can be modified by setting either the `[tool.slap.plugins.disable]` or `[tool.slap.plugins.enable]`
        configuration option (without the `tool.slap` prefix in case of a `slap.toml` configuration file). The default
        plugins delivered immediately with Slap are enabled by default unless disabled explicitly with the `disable`
        option."""

        from nr.util.plugins import iter_entrypoints

        from slap.plugins import ApplicationPlugin

        assert not self._plugins_loaded
        self._plugins_loaded = True

        config = self.config()
        disable = config.disable or []

        logger.debug("Loading application plugins")

        for plugin_name, loader in iter_entrypoints(ApplicationPlugin):  # type: ignore[misc]
            if plugin_name in disable:
                continue
            try:
                plugin = loader()()
            except Exception:
                logger.exception("Could not load plugin <subj>%s</subj> due to an exception", plugin_name)
            else:
                plugin_config = plugin.load_configuration(self)
                plugin.activate(self, plugin_config)

    def _cleo_init(self, io: IO) -> None:
        self.load_plugins()

    def run(self) -> None:
        """Loads and activates application plugins and then invokes the CLI."""

        self.cleo.run()



from __future__ import annotations

import dataclasses
import logging
import typing as t
from pathlib import Path

from databind.core.settings import Alias

from slap.configuration import Configuration

if t.TYPE_CHECKING:
    from nr.util.functional import Once

    from slap.install.installer import Indexes
    from slap.plugins import ProjectHandlerPlugin
    from slap.python.dependency import Dependency, VersionSpec
    from slap.release import VersionRef
    from slap.repository import Repository


logger = logging.getLogger(__name__)


@dataclasses.dataclass
class Dependencies:
    python: VersionSpec | None
    run: t.Sequence[Dependency]
    dev: t.Sequence[Dependency]
    extra: t.Mapping[str, t.Sequence[Dependency]]
    build: t.Sequence[Dependency]
    indexes: Indexes = None  # type: ignore  # To avoid having to import the Indexes class globally

    def __post_init__(self) -> None:
        from slap.install.installer import Indexes

        if self.indexes is None:
            self.indexes = Indexes()  # type: ignore[unreachable]


@dataclasses.dataclass
class Package:
    name: str  #: The name of the package. Contains periods in case of a namespace package.
    path: Path  #: The path to the package directory. This points to the namespace package if applicable.
    root: Path  #: The root directory that contains the package.


@dataclasses.dataclass
class ProjectConfig:
    #: The name of the project handler plugin. If none is specified, the built-in project handlers are tested
    #: (see the #slap.ext.project_handlers module for more info on those).
    handler: str | None = None

    #: The source directory to use when relying on automatic package detection. If not set, the default project
    #: handler will search in `"src/"`` and then `"./"``.
    source_directory: t.Annotated[str | None, Alias("source-directory")] = None

    #: Whether the project source code is intended to be typed.
    typed: bool | None = None


class Project(Configuration):
    """Represents one Python project. Slap can work with multiple projects at the same time, for example if the same
    repository or source code project contains multiple individual Python projects. Every project has its own
    configuration, either loaded from `slap.toml` or `pyproject.toml`."""

    #: Reference to the Slap application object.
    repository: Repository

    #: The parsed configuration, accessible as a #Once.
    config: Once[ProjectConfig]

    #: The packages detected with #get_packages() as a #Once.
    packages: Once[t.Sequence[Package] | None]

    #: The packages detected readme as a #Once.
    readme: Once[str | None]

    #: The packages dependencies as a #Once.
    dependencies: Once[Dependencies]

    def __init__(self, repository: Repository, directory: Path) -> None:
        super().__init__(directory)
        from nr.util.functional import Once

        from slap.util.toml_file import TomlFile

        self.repository = repository
        self.usercfg = TomlFile(Path("~/.config/slap/config.toml").expanduser())
        self.handler = Once(self._get_project_handler)
        self.config = Once(self._get_project_configuration)
        self.packages = Once(self._get_packages)
        self.readme = Once(self._get_readme)
        self.dependencies = Once(self._get_dependencies)
        self.dist_name = Once(self._get_dist_name)
        self.version = Once(self._get_version)

    def _get_project_configuration(self) -> ProjectConfig:
        """Loads the project-level configuration."""

        from databind.core.settings import ExtraKeys
        from databind.json import load

        return load(self.raw_config(), ProjectConfig, settings=[ExtraKeys(True)])

    def _get_project_handler(self) -> ProjectHandlerPlugin:
        """Returns the handler for this project."""

        from nr.util.plugins import iter_entrypoints, load_entrypoint

        from slap.plugins import ProjectHandlerPlugin

        handler_name = self.config().handler
        if handler_name is None:
            for handler_name, loader in iter_entrypoints(ProjectHandlerPlugin):  # type: ignore[misc]
                handler = loader()()
                if handler.matches_project(self):
                    break
            else:
                raise RuntimeError(f"unable to identify project handler for {self!r}")
        else:
            assert isinstance(handler_name, str), repr(handler_name)
            handler = load_entrypoint(ProjectHandlerPlugin, handler_name)()  # type: ignore[misc]
            assert handler.matches_project(self), (self, handler)
        return handler

    def _get_packages(self) -> list[Package] | None:
        """Returns the packages that can be detected for this project. How the packages are detected depends on the
        #ProjectConfig.packages option."""

        if not self.is_python_project:
            return []
        packages = self.handler().get_packages(self)
        if packages:
            logger.debug(
                "Detected packages for project <subj>%s</subj> by package detector <obj>%s</obj>: <val>%s></val>",
                self,
                self.handler(),
                packages,
            )
        elif self.is_python_project and packages is not None:
            logger.warning(
                "No packages detected for project <subj>%s</subj> by any of package detectors <val>%s</val>",
                self,
                self.handler(),
            )
        return packages

    def _get_dist_name(self) -> str | None:
        return self.handler().get_dist_name(self)

    def _get_readme(self) -> str | None:
        return self.handler().get_readme(self)

    def _get_dependencies(self) -> Dependencies:
        return self.handler().get_dependencies(self)

    def _get_version(self) -> str | None:
        return self.handler().get_version(self)

    def get_version_refs(self) -> list[VersionRef]:
        return self.handler().get_version_refs(self)

    def get_interdependencies(self, projects: t.Sequence[Project], recursive: bool = False) -> list[Project]:
        """Returns the dependencies of this project in the list of other projects. The returned dictionary maps
        to the project and the dependency constraint. This will only take run dependencies into account."""

        dependency_names = set()
        for dep in self.dependencies().run:
            dependency_names.add(dep.name)

        result = []
        for project in projects:
            if project.dist_name() in dependency_names:
                result.append(project)
                if recursive:
                    result += project.get_interdependencies(projects, True)

        return result

    def add_dependency(self, dependency: Dependency, where: str) -> None:
        """Add a dependency to the project configuration.

        Arguments:
          selector: The dependency to add.
          where: The location of where to add the dependency. This is either `'run'`, `'dev'`, or otherwise
            refers to the name of an extra requirement.
        Raises:
          NotImplementedError: If the operation is not supported on the project.
        """

        from slap.python.dependency import Dependency

        assert isinstance(dependency, Dependency), type(dependency)
        self.handler().add_dependency(self, dependency, where)
        # TODO(@NiklasRosenstein): Use a method to flush the cache of Once when it is available in `nr.utils`.
        self.raw_config.get(True)
        self.dependencies.get(True)

    @property
    def id(self) -> str:
        return self.dist_name() or self.directory.resolve().name

    @id.setter
    def id(self, value: str) -> None:
        self._id = value

    @property
    def is_python_project(self) -> bool:
        return self.pyproject_toml.exists()



import dataclasses
import typing as t

from databind.core.settings import Alias
from nr.util.fs import get_file_in_directory

from slap.plugins import RepositoryHandlerPlugin
from slap.project import Project
from slap.repository import Repository, RepositoryHost
from slap.util.vcs import Vcs, detect_vcs


@dataclasses.dataclass
class DefaultRepositoryConfig:
    #: A list of paths pointing to projects to include in the application invokation. This is useful if multiple
    #: projects should be usable with the Slap CLI in unison. Note that if this option is not set and either no
    #: configuration file exists in the CWD or the `slap.toml` is used, all immediate subdirectories that contain
    #: a `pyproject.toml` will be considered included projects.
    include: list[str] | None = None

    #: The repository hosting service. If not specified, it will be detected automatically.
    repository_host: t.Annotated[RepositoryHost | None, Alias("repository-host")] = None


class DefaultRepositoryHandler(RepositoryHandlerPlugin):
    """The default implementation of the repository handler.

    Applies only if either

    * A VCS can be detected (and projects are loaded from the VCS root).
    * If a README file, LICENSE file, or a Pyproject or Slap configuration file exists. This is to avoid mistakenly
      considering a directory that contains independent Python projects as a mono-repository.

    !!! note

        In a future version, this handler may update the #Repository.directory to point to the VCS root directory
        (if a VCS can be detected) to allow using the Slap CLI from a subdirectory as if it were used in the root
        directory.
    """

    def _get_config(self, repository: Repository) -> DefaultRepositoryConfig:
        import databind.json

        raw_config = repository.raw_config().get("repository", {})
        raw_config.pop("handler", None)
        config = databind.json.load(raw_config, DefaultRepositoryConfig)
        return config

    def matches_repository(self, repository: Repository) -> bool:
        if repository.pyproject_toml.exists() or repository.slap_toml.exists():
            return True
        # NOTE(@NiklasRosenstein): This is where we would update the repository root directory.
        # vcs = self.get_vcs(repository)
        # if vcs is not None:
        #   repository.__init__(vcs.get_toplevel())
        #   return True
        if get_file_in_directory(repository.directory, "readme", [], case_sensitive=False):
            return True
        if get_file_in_directory(repository.directory, "license", [], case_sensitive=False):
            return True
        return False

    def get_vcs(self, repository: Repository) -> Vcs | None:
        return detect_vcs(repository.directory)

    def get_repository_host(self, repository: Repository) -> RepositoryHost | None:
        from nr.util.plugins import iter_entrypoints

        config = self._get_config(repository)
        if config.repository_host:
            return config.repository_host
        for _plugin_name, loader in iter_entrypoints(RepositoryHost):  # type: ignore[misc]
            if instance := loader().detect_repository_host(repository):
                return instance
        return None

    def get_projects(self, repository: Repository) -> list[Project]:
        from slap.project import Project

        projects = []
        if repository.pyproject_toml.exists():
            projects.append(Project(repository, repository.directory))

        config = self._get_config(repository)
        if config.include is None or not repository.pyproject_toml.exists():
            for path in repository.directory.iterdir():
                if not path.is_dir():
                    continue
                project = Project(repository, path)
                if project.pyproject_toml.exists():
                    projects.append(project)
        else:
            for subdir in config.include:
                projects.append(Project(repository, repository.directory / subdir))

        return projects



import dataclasses
import io
import logging
import re
import typing as t
from pathlib import Path

from databind.core.settings import Alias, ExtraKeys

from slap.application import Application, Command, argument, option
from slap.changelog import Changelog, ChangelogManager, ManagedChangelog
from slap.plugins import ApplicationPlugin, ChangelogUpdateAutomationPlugin
from slap.project import Project
from slap.repository import Issue, PullRequest, Repository
from slap.util.pygments import toml_highlight

logger = logging.getLogger(__name__)

#: The default set of types a changelog entry can have.
DEFAULT_VALID_TYPES = [
    "breaking change",
    "deprecation",
    "docs",
    "feature",
    "fix",
    "hygiene",
    "improvement",
    "refactor",
    "tests",
]


def get_default_author(app: Application) -> str | None:
    vcs = app.repository.vcs()
    remote = app.repository.host()
    return remote.get_username(app.repository) if remote else vcs.get_author().email if vcs else None


@ExtraKeys(True)
@dataclasses.dataclass
class ChangelogConfig:
    #: Whether the changelog feature is enabled. This acts locally for the current project and not globally.
    #: Particularly useful for monorepos that have multiple subprojects each with their changelog directories
    #: to prevent accidentally creating changelogs in the root directory.
    #:
    #: When not set, it will be considered `True` if the current project is a Python project.
    enabled: bool | None = None

    #: The directory in which changelogs are stored.
    directory: Path = Path(".changelog")

    #: The list of valid types that can be used in changelog entries. The default types are
    #: #DEFAULT_CHANGELOG_TYPES.
    valid_types: t.Annotated[list[str] | None, Alias("valid-types")] = dataclasses.field(
        default_factory=lambda: list(DEFAULT_VALID_TYPES)
    )


class BaseChangelogCommand(Command):
    def __init__(self, app: Application, manager: ChangelogManager) -> None:
        super().__init__()
        self.app = app
        self.manager = manager


class ChangelogAddCommand(BaseChangelogCommand):
    """Add an entry to the unreleased changelog via the CLI.

    A changelog is a TOML file, usually in the <u>.changelog/</u> directory, named with
    the version number it refers to and containing changelog entries. Changes that
    are currently not released in a version are stored in a file called
    <u>_unreleased.toml</u>.

    Changelog entries contain at least one author, a type (e.g. whether the entry
    describes a feature, enhancement, bug fix, etc.) and optionally a subject (e.g.
    whether the change is related to docs or a particular component of the code), a
    Markdown description, possibly a link to a pull request with which the change
    was introduced and links to issues that the changelog addresses.

    <b>Example:</b>

      <fg=blue># .changelog/0.1.1.toml</fg>
      <fg=cyan>[changelog]</fg>
      <fg=green>release-date</fg> = <fg=yellow>"2022-01-17"</fg>

      <fg=cyan>[[changelog.entries]]</fg>
      <fg=green>id</fg> = <fg=yellow>"a7bc01f"</fg>
      <fg=green>type</fg> = <fg=yellow>"improvement"</fg>
      <fg=green>description</fg> = <fg=yellow>"Improvement to `my_package.util`"</fg>
      <fg=green>author</fg> = <fg=yellow>"username"</fg>
      <fg=green>pr</fg> = <fg=yellow>"https://github.com/username/my_package/pulls/13"</fg>

    Changelog entries can be managed easily using the <info>slap log</info> command.

      <fg=yellow>$</fg> slap log add -t feature -d 'Improvement to `my_package.util`"

    The <fg=green>pr</fg> field is usually set manually after the PR is created or updated
    automatically by a CI action using the <info>slap log update-pr-field</info> command.
    """

    name = "changelog add"

    options = [
        option(
            "--type",
            "-t",
            description="The type of the changelog. Unless configured differently, one of "
            "{', '.join(DEFAULT_VALID_TYPES)}",
            flag=False,
        ),
        option(
            "--description",
            "-d",
            description="A Markdown formatted description of the changelog entry.",
            flag=False,
        ),
        option(
            "--author",
            "-a",
            description="Your username or email address. By default, this will be your configured "
            "Git name and email address.",
            flag=False,
        ),
        option(
            "--pr",
            None,
            description="The pull request that the change is introduced to the main branch with. This is not usually "
            "known at the time the changelog entry is created, so this option is not often used. If the remote "
            "repository is well supported by Slap, a pull request number may be specified and converted to a full "
            "URL by Slap, otherwise a full URL must be specified.",
            flag=False,
        ),
        option(
            "--issue",
            "-i",
            description="An issue related to this changelog. If the remote repository is well supported by Slap, "
            "an issue number may be specified and converted to a full URL by Slap, otherwise a full URL must be "
            "specified.",
            flag=False,
            multiple=True,
        ),
        option(
            "--commit",
            "-c",
            description="Commit the currently staged changes in the VCS as well as the updated changelog file to disk. "
            "The commit message is a concatenation of the <opt>--type, -t</opt> and <opt>--description, -d</opt>, as "
            "well as the directory relative to the VCS toplevel if the changelog is created not in the toplevel "
            "directory of the repository.",
        ),
    ]

    def handle(self) -> int:
        if self.manager.readonly:
            self.line_error("error: cannot add changelog because the feature must be enabled in the config", "error")
            return 1

        vcs = self.app.repository.vcs()
        change_type: str | None = self.option("type")
        description: str | None = self.option("description")
        author: str | None = self.option("author") or get_default_author(self.app)
        pr: str | None = self.option("pr")
        issues: list[str] | None = self.option("issue")

        if not vcs and self.option("commit"):
            self.line_error("error: no VCS detected, but <opt>--commit, -c</opt> was used", "error")
            return 1

        if not change_type:
            self.line_error("error: missing <opt>--type,-t</opt>", "error")
            return 1
        if not description:
            self.line_error("error: missing <opt>--description,-d</opt>", "error")
            return 1
        if not author:
            self.line_error("error: missing <opt>--author,-a</opt>", "error")
            return 1

        entry = self.manager.make_entry(change_type, description, author, pr, issues)
        unreleased = self.manager.unreleased()
        changelog = unreleased.content if unreleased.exists() else Changelog()
        changelog.entries.append(entry)
        unreleased.save(changelog)

        print(toml_highlight(self.manager.deser.dump_entry(entry)))

        if self.option("commit"):
            assert vcs is not None
            commit_message = f"{change_type}: {description}"
            main_project = self.app.main_project()
            relative = main_project.directory.relative_to(self.app.repository.directory) if main_project else Path(".")
            if relative != Path("."):
                prefix = str(relative).replace("\\", "/").strip("/")
                commit_message = f"{prefix}/: {commit_message}"
            vcs.commit_files([unreleased.path], commit_message)

        return 0


class ChangelogUpdatePrCommand(Command):
    """Update the <code>pr</code> field of changelog entries in a commit range.

    Updates all changelog entries that were added in a given commit range. This is
    useful to run in CI for a pull request to avoid having to manually update the
    changelog entry after the PR has been created.
    """

    name = "changelog update-pr"
    arguments = [
        argument(
            "base_revision",
            description="The revision ID to look back to to make out which changelog entries have been added since.",
            optional=True,
        ),
        argument(
            "pr",
            description="The reference to the PR that should be inserted into all entries added between the specified "
            "revision and the current version of the unreleased changelog.",
            optional=True,
        ),
    ]
    options = [
        option(
            "--dry",
            "-d",
            description="Do not actually make changes on disk.",
        ),
        option(
            "--overwrite",
            description="Update PR references even if an entry's reference is already set but different.",
        ),
        option(
            "--commit",
            "-c",
            description="Commit the changes, if any.",
        ),
        option(
            "--push",
            "-p",
            description="Push the changes, if any.",
        ),
        option(
            "--name",
            description="Override the <code>user.name</code> Git option (only with <opt>--commit, -c</opt>)",
            flag=False,
        ),
        option(
            "--email",
            description="Override the <code>user.email</code> Git option (only with <opt>--commit, -c</opt>).",
            flag=False,
        ),
        option(
            "--use",
            description="Use the specified plugin to publish the updated changelogs. Use this in supported CI "
            "environments instead of manually configuring the command-line settings.",
            flag=False,
        ),
        option(
            "--list",
            "-l",
            description="List the available plugins you can pass to the <opt>--use</opt> option.",
        ),
    ]

    def __init__(self, app: Application):
        super().__init__()
        self.app = app
        self.managers = {
            config: get_changelog_manager(app.repository, config if isinstance(config, Project) else None)
            for config in app.configurations()
        }

    def handle(self) -> int:
        from nr.util.plugins import iter_entrypoints, load_entrypoint

        if not self._validate_arguments():
            return 1

        if self.option("list"):
            for ep in iter_entrypoints(ChangelogUpdateAutomationPlugin.ENTRYPOINT):
                self.line(f"  • {ep.name}")
            return 0

        automation_plugin: ChangelogUpdateAutomationPlugin | None = None
        if plugin_name := self.option("use"):
            logger.info("Loading changelog update automation plugin <subj>%s</subj>", plugin_name)
            automation_plugin = load_entrypoint(ChangelogUpdateAutomationPlugin, plugin_name)()  # type: ignore[misc]
            automation_plugin.io = self.io
            automation_plugin.initialize()
            base_revision: str = automation_plugin.get_base_ref()
        else:
            base_revision = self.argument("base_revision")
        assert base_revision

        changelogs: list[tuple[ManagedChangelog, ChangelogManager]] = []
        for _, manager in self.managers.items():
            unreleased = manager.unreleased()
            if unreleased.exists():
                changelogs.append((unreleased, manager))

        if not changelogs:
            self.line("no entries to update", "info")
            return 0

        if not (pr := self.argument("pr")):
            assert automation_plugin
            pr = automation_plugin.get_pr()

        host = self.app.repository.host()
        if host:
            try:
                pr = host.get_pull_request_by_reference(pr).url
            except ValueError as exc:
                self.line_error(f"error: {exc}", "error")
                return 1

        num_updates = 0
        for changelog, manager in changelogs:

            prev_contents = self._vcs.get_file_contents(changelog.path, base_revision)
            if prev_contents is None:
                prev_entry_ids = set[str]()
            else:
                prev_changelog = manager.deser.load(
                    io.StringIO(prev_contents.decode("utf8")), f"{base_revision}:{changelog.path}"
                )
                prev_entry_ids = {e.id for e in prev_changelog.entries}

            new_entry_ids = {e.id for e in changelog.content.entries if e.pr != pr} - prev_entry_ids
            entries_to_update = [
                e for e in changelog.content.entries if e.id in new_entry_ids and (not e.pr or self.option("overwrite"))
            ]
            if not entries_to_update:
                continue

            self.line(
                f"update <info>{changelog.path.relative_to(Path.cwd())}</info> "
                f'({len(entries_to_update)} reference{"s" if len(new_entry_ids) != 1 else ""})'
            )

            num_updates += len(entries_to_update)
            for entry in entries_to_update:
                entry.pr = pr

            if not self.option("dry"):
                changelog.save(None)

        if not num_updates:
            self.line("no entries to update", "info")
            return 0

        if self.option("dry"):
            return 0

        changed_files = [changelog.path for changelog, _ in changelogs]

        if self.option("commit"):
            assert not self.automation_plugin
            self._vcs.commit_files(
                changed_files,
                f'Updated {num_updates} PR reference{"" if num_updates == 0 else "s"}.',
                push=self._remote,
                name=self.option("name"),
                email=self.option("email"),
            )

        if automation_plugin:
            automation_plugin.publish_changes(changed_files)

        return 0

    def _validate_arguments(self) -> bool:
        if self.option("list"):
            if used_option := next((o.name for o in self.options if o.name != "list" and self.option(o.name)), None):
                self.line_error(f"error: <opt>--{used_option}</opt> cannot be used with <opt>--list</opt>", "error")
                return False

        elif self.option("use"):
            if used_option := next((o.name for o in self.options if o.name != "use" and self.option(o.name)), None):
                self.line_error(f"error: <opt>--{used_option}</opt> cannot be used with <opt>--use</opt>", "error")
                return False

        else:
            for arg in ("base_revision", "pr"):
                if not self.argument(arg):
                    self.line_error(f"error: missing argument <opt>{arg}</opt>", "error")
                    return False

        if self.option("push") and not self.option("commit"):
            self.line_error(
                "error: <opt>--push, -p</opt> can only be used in combination with <opt>--commit, -c</opt>", "error"
            )
            return False

        vcs = self.app.repository.vcs()
        if not vcs:
            self.line_error("error: VCS is not configured or could not be detected", "error")
            return False
        self._vcs = vcs

        if self.option("push"):
            self._remote = next((r for r in self._vcs.get_remotes() if r.default), None)
        else:
            self._remote = None

        for opt in ("email", "name"):
            if self.option(opt) and not self.option("commit"):
                self.line_error(f"error: <opt>--{opt}</opt> is not valid without <opt>--commit, -c</opt>", "error")
                return False

        return True


class ChangelogFormatCommand(BaseChangelogCommand):
    """Format the changelog in the terminal or in Markdown format."""

    name = "changelog format"
    options = [
        option(
            "--markdown",
            "-m",
            description="Render the changelog in Markdown format.",
        ),
        option(
            "--all",
            "-a",
            description="Render all changelogs in reverse chronological order.",
        ),
    ]
    arguments = [
        argument("version", "The changelog version to format.", optional=True),
    ]

    def handle(self) -> int:
        if not self._validate_arguments():
            return 1

        if self.option("all"):
            changelogs = self.manager.all()
        elif version := self.argument("version"):
            changelogs = [self.manager.version(version)]
            if not changelogs[0].exists():
                self.line_error(f'error: Changelog for <opt>version</opt> "{version}" does not exist.', "error")
                return 1
        else:
            changelogs = [self.manager.unreleased()]

        for changelog in changelogs:
            if self.option("markdown"):
                self._render_markdown(changelog)
            else:
                self._render_terminal(changelog)
            self.line("")

        return 0

    def _validate_arguments(self) -> bool:
        if self.option("all") and self.argument("version"):
            self.line_error("error: <opt>--all, -a</opt> is incompatible with <opt>version</opt> argument", "error")
            return False
        return True

    def _render_terminal(self, changelog: ManagedChangelog) -> None:
        if changelog.version:
            assert changelog.content.release_date
            self.line(f"<b>{changelog.version}</b> (<u>{changelog.content.release_date}</u>)")
        else:
            self.line(f'<b>{changelog.version or "Unreleased"}</b>')
            if not changelog.exists():
                return

        for entry in changelog.content.entries:
            description = entry.description
            description = re.sub(r"(?!<\\)`([^`]+)`", r"<fg=dark_gray>\1</fg>", description)
            self.line(f"  <fg=cyan;options=italic>{entry.type}</fg> — {description} (<fg=yellow>{entry.author}</fg>)")

    def _render_markdown(self, changelog: ManagedChangelog) -> None:
        if changelog.version:
            assert changelog.content.release_date
            print(f"## {changelog.version} ({changelog.content.release_date})")
        else:
            print("## Unreleased")
            if not changelog.exists():
                return

        print()
        print("<table><tr><th>Type</th><th>Description</th><th>PR</th><th>Issues</th><th>Author</th></tr>")
        for entry in changelog.content.entries:
            pr_link = self._html_anchor("pr", entry.pr) if entry.pr else ""
            issues = ", ".join(self._html_anchor("issue", issue) for issue in entry.issues) if entry.issues else ""
            print(
                f"  <tr><td>{entry.type.capitalize()}</td><td>\n\n{entry.description}</td>"
                f'<td>{pr_link}</td><td>{issues}</td><td>{", ".join(entry.get_authors())}</td></tr>'
            )
        print("</table>")

    def _html_anchor(self, type: t.Literal["pr", "issue"], ref: str) -> str:
        if self.manager.repository_host:
            try:
                if type == "pr":
                    item: Issue | PullRequest = self.manager.repository_host.get_pull_request_by_reference(ref)
                else:
                    item = self.manager.repository_host.get_issue_by_reference(ref)
            except ValueError as exc:
                self.line_error(f"warning: {exc}", "error")
                url = ref
                text = "Link"
            else:
                url, text = item.url, item.id
        return f'<a href="{url}">{text}</a>'


class ChangelogConvertCommand(BaseChangelogCommand):
    """Convert Slap's old YAML based changelogs to new style TOML changelogs.

    Sometimes the changelog entries in the old style would be suffixed with the
    author's username in the format of <code>@Name</code> or <code>(@Name)</code>, so this command will
    attempt to extract that information to figure out the author of the change.
    """

    name = "changelog convert"
    options = [
        option(
            "--author",
            "-a",
            description="The author to fall back to. If not specified, the current VCS queried for the "
            "author name instead and their email will be used (depending on the normalization of the "
            "repository remote, this will be converted to a username, for example in the case of GitHub).",
            flag=False,
        ),
        option(
            "--directory",
            "-d",
            description="The directory from which to load the old changelogs. Defaults to the same directory that the "
            "new changelogs will be written to.",
            flag=False,
        ),
        option("--dry", description="Do not make changes on disk."),
        option(
            "--fail-fast",
            "x",
            description="If this flag is enabled, exit as soon as an error is encountered with any file.",
        ),
    ]

    CHANGELOG_TYPE_MAPPING_TABLE = {
        "change": "improvement",
        "break": "breaking change",
        "breaking_change": "breaking change",
        "refactor": "hygiene",
    }

    def handle(self) -> int:
        import yaml

        # vcs = self.app.repository.vcs()
        author = self.option("author") or get_default_author(self.app)

        if not author:
            self.line_error("error: missing <opt>--author,-a</opt>", "error")
            return 1

        directory = self.option("directory") or self.manager.directory
        has_failures = False
        for filename in directory.iterdir():
            if has_failures and self.option("fail-fast"):
                break
            if filename.suffix in (".yaml", ".yml"):
                try:
                    self._convert_changelog(author, filename)
                except yaml.error.YAMLError as exc:
                    has_failures = True
                    self.line_error(f'warn: cannot parse "{filename}": {exc}', "warning")
                    continue
                except Exception as exc:
                    has_failures = True
                    self.line_error(f'warn: could not convert "{filename}": {exc}', "warning")
                    if self.io.is_very_verbose:
                        import traceback

                        self.line_error(traceback.format_exc())
                    continue

        return 1 if has_failures else 0

    def _convert_changelog(self, default_author: str, source: Path) -> None:
        import datetime

        import yaml

        data = yaml.safe_load(source.read_text())
        entries = []
        for original_entry in data["changes"]:
            prefix = ""
            component = original_entry["component"]
            if component == "docs":
                change_type = "docs"
            elif component in ("test", "tests"):
                change_type = "tests"
            else:
                change_type = original_entry["type"]
                prefix = f"{component}: " if component != "general" else ""
            author, original_entry["description"] = self._match_author_in_description(original_entry["description"])
            new_entry = self.manager.make_entry(
                change_type=self.CHANGELOG_TYPE_MAPPING_TABLE.get(change_type, change_type),
                description=prefix + original_entry["description"],
                author=author or default_author,
                pr=None,
                issues=original_entry.get("fixes", None) or None,
            )
            entries.append(new_entry)

        if source.stem == "_unreleased":
            dest = self.manager.unreleased()
        else:
            dest = self.manager.version(source.stem)

        changelog = dest.content if dest.exists() else Changelog()
        changelog.release_date = (
            datetime.datetime.strptime(data["release_date"], "%Y-%m-%d").date() if data.get("release_date") else None
        )
        changelog.entries = entries

        if self.option("dry"):
            self.io.write_line(f"<fg=cyan;options=underline># {dest.path}</fg>")
            print(toml_highlight(self.manager.deser.dump(changelog)))
        else:
            dest.save(changelog)

    def _match_author_in_description(self, description: str) -> tuple[str | None, str]:
        """Internal. Tries to find the @Author at the end of a changelog entry description."""

        import re

        match = re.search(r"(.*)\((@[\w\-_ ]+)\)$", description)
        return match.group(2) if match else None, match.group(1).strip() if match else description


class ChangelogCommandPlugin(ApplicationPlugin):
    def load_configuration(self, app: Application) -> ChangelogManager:
        return get_changelog_manager(app.repository, app.main_project())

    def activate(self, app: "Application", config: ChangelogManager) -> None:
        app.cleo.add(ChangelogAddCommand(app, config))
        app.cleo.add(ChangelogUpdatePrCommand(app))
        app.cleo.add(ChangelogFormatCommand(app, config))
        app.cleo.add(ChangelogConvertCommand(app, config))


def get_changelog_manager(repository: Repository, project: Project | None) -> ChangelogManager:
    import databind.json

    config = databind.json.load((project or repository).raw_config().get("changelog", {}), ChangelogConfig)
    if config.enabled is None and project:
        config.enabled = project.is_python_project

    return ChangelogManager(
        directory=(project or repository).directory / config.directory,
        repository_host=repository.host(),
        valid_types=config.valid_types,
        readonly=not config.enabled,
    )
